/*+++++++++++++++       
.TYPE           Module
.IDENTIFICATION coos.c
.VERSION 1.0	24-Oct-1988: Creation
.VERSION 1.1	05-May-1989: Corrected bug in SuperGalactic
.VERSION 2.0	25-Feb-1992: Tested on PC, w. Turbo-C
.VERSION 2.1	07-Nov-1994: Keep the precision...
.VERSION 3.0	01-Jul-1998: Take care of errors, and ICRS System.
.VERSION 3.1	21-Jul-1998: tr_s2 returns _COO_EST_xxx flags
.VERSION 3.2	02-Dec-1998: take care of absent / zero parallax
.VERSION 3.3	18-Mar-1999: tr_setframe
.VERSION 3.4	30-Aug-2002: tr_setframe initialize epoch = epocho
.VERSION 3.41	03-Apr-2003: Compute default errors for proper motions
.VERSION 3.42	22-Nov-2004: Take care u==0 (no position)
				(syndrome: RV becomes zero...)
.VERSION 3.43	28-Aug-2010: Take care no velocity
.VERSION 3.44	16-Oct-2013: Circular error ellipse ==> keep theta=0
.LANGUAGE       C
.AUTHOR         P.T Wallace [Starlink], Francois Ochsenbein [ESO-IPG]  
.KEYWORDS       Change of Coordinates

.ENV            

.COMMENTS
This module contains routines to transform coordinates
from /to one of the systems FK4, FK5, ICRS, Galactic, Ecliptic, Supergalactic.

\begin{TeX}

The modification concerns the position as well as the
velocity vector (radial velocity and proper motions).
The complete position information is stored in a 
`s' structure containing the 6--vector $\{x y z \dot{x} ,\dot{y} \dot{z}\}$,
the epoch, the equinox, and the coordinate system, as well as uncertainties
(covariances matrix)

The velocity vector $\vec{v}(t)$ contains the velocity expressed in $''$/year 
as:

$$\begin{array}{ll}
  v_0 = k \pi V_{r0} 		& \\
  v_1 = \mu_{\alpha0} \cos \delta_0 &  	\\
  v_2 = \mu_{\delta0} 		&  	
\end{array}$$

The value of $k$  ($0.21094502$) when $\pi$ is
expressed in " and $V_r$ in km/s
is defined in {\tt ASTRO.H} as the mnemonic {\tt km\_s\_pc}.

\underline{For ICRS system}, use Hipparcos Conventions, with 
	parallax in mas, proper motions in mas/yr.

The error computations use the $6\times 6$ covariance matrices: the basic used
theorem is the multinormal distribution, which specifies that:
if \vec{x} is multinormal with covariance matrix {\bf V}, then
\vec{y} = {\bf B}\vec{x} is multinormal with covariance matrix
{\bf $B . V . ^tB$}

The error propagation is computed from covariance matrices 
with the transformation based on 6D matrix
$$T(t) = \begin{array}{rrrrrr}
	1 & 0 & 0 & t & 0 & 0 \\
	0 & 1 & 0 & 0 & t & 0 \\
	0 & 0 & 1 & 0 & 0 & t \\
	0 & 0 & 0 & 1 & 0 & 0 \\
	0 & 0 & 0 & 0 & 1 & 0 \\
	0 & 0 & 0 & 0 & 0 & 1 \\
      \end{array}
applied on 6-vector $(\vec{u}, \vec{\dot{u}}). Note that such matrices
have  $T^-1(t) = T(-t)$ , and $T(t_1) \cdot T(t_2) = T(t_1+t_2) $

Errors may also be expressed in terms of {\em error ellipses}:
the error ellipse has the position angle theta (from North through East),
and non-correlated semi-major and semi-minor axises a and b .
\end{TeX}

----------------------*/

#include <astropos.h>
#include <stdio.h>
#include <trigo.h>
#include <string.h>	/* memset */
#ifndef VOID
#define VOID void
#endif
#define ITEMS(a)	(sizeof(a)/sizeof(a[0]))

static double R44[3][3], y40 = -99999999.e0, y41 = -99999999.e0;
static double R55[3][3], y50 = -99999999.e0, y51 = -99999999.e0;
static double R45[3][3];
static double R54[3][3];

static double RB0[3][3], yB0 = -99999999.e0;
static double RB1[3][3], yB1 = -99999999.e0;

static double RJ0[3][3], yJ0 = -99999999.e0;
static double RJ1[3][3], yJ1 = -99999999.e0;

static double RE0[3][3], yE0 = -99999999.e0;
static double RE1[3][3], yE1 = -99999999.e0;

static double Rl[3][3] = {	{1.e0, 0.e0, 0.e0},	/* xyz into Local */
				{0.e0, 1.e0, 0.e0},
				{0.e0, 0.e0, 1.e0}};
			
static double R6[6][6] ;

static double GAL_2000[3][3] = {	/* J2000 to Supergalactic	*/
	{1.e0, 0.e0, 0.e0},
	{0.e0, 1.e0, 0.e0},
	{0.e0, 0.e0, 1.e0}};

extern double gal_2000[3][3];		/* J2000 to galactic (jgal.c)	*/
extern double supergal[3][3];		/* Galactic to Supergalactic	
						(supergal.c)		*/

/* Default error ellipse for position */
#define e_1pos	2.1	/* Default error on one position = 3/sqrt(2)	*/
static double def_err[] = { 	/* Errors in mas	*/
    180.*3.6e6,   e_1pos*3.6e6, e_1pos*3.6e5,		/* deg / 0.1deg */
    e_1pos*6.e4 , e_1pos*6.e3,  e_1pos*1000., 		/* ' .1' "	*/
    e_1pos*100. , e_1pos*10.  , e_1pos      , 		/* 100 10 1 mas	*/
    e_1pos/10.  , e_1pos/100. , e_1pos/1000. 
  } ;
static double def_rv[2]  = {0.,50.} ;	/* Default radial velocity	*/
static double def_plx[2] = {9.,10.} ;	/* Default parallax in mas	*/
static double zero_plx   = 0.100 ;	/* Max distance = 10kpc		*/

#define scalar(a,b)	(a[0]*b[0]+a[1]*b[1]+a[2]*b[2])

/*============================================================================*/

#ifdef TEST
#include <stdio.h>
static VOID pr3(char *title, char *fmt, double v[3])
{
    printf("%s: ", title) ;
    printf(fmt, v[0]); putchar(' ') ;
    printf(fmt, v[1]); putchar(' ') ;
    printf(fmt, v[2]); putchar('\n') ;
}
static VOID pr4(char *title, char *fmt, double v[3], double x)
{
    printf("%s: ", title) ;
    printf(fmt, v[0]); putchar(' ') ;
    printf(fmt, v[1]); putchar(' ') ;
    printf(fmt, v[2]); putchar(' ') ;
    printf("%6.1fkm/s", x); putchar('\n') ;
}
static VOID pr6(char *title, char *fmt, double v[6])
{
    printf("%s: ", title) ;
    printf(fmt, v[0]); putchar(' ') ;
    printf(fmt, v[1]); putchar(' ') ;
    printf(fmt, v[2]); putchar(' ') ;
    printf(fmt, v[3]); putchar(' ') ;
    printf(fmt, v[4]); putchar(' ') ;
    printf(fmt, v[5]); putchar('\n') ;
}
static VOID pr66(char *title, char *fmt, double m[6][6])
{
  int i ;
    printf("%s\n", title) ;
    for (i=0; i<6; i++) pr6("\t", fmt, m[i]) ;
}
#else
#define pr3(t,f,v)
#define pr4(t,f,v,x)
#define pr6(t,f,v)
#define pr66(t,f,v)
#endif

/*============================================================================*
 *			Computation of variance/covariance
 *============================================================================*/
int tr_e3var (e3, v3)
/*+++
.PURPOSE   Transformation of error ellipse into (variance, covariance)
.RETURNS   0 
.REMARKS   e3[0] = a,      e3[1] = b,      e3[2] = theta ; 
	   v3[0] = var(x), v3[1] = var(y), v3[2] = cov(x,y)
---*/
  double e3[3] ;	/* IN: Error ellipse	*/
  double v3[2] ;	/* OUT: Variances + cov	*/
{
  double st, ct, a2, b2 ;
    a2 = e3[0] * e3[0] ; b2 = e3[1] * e3[1] ;
    ct = cosd(e3[2]) ;   st = sind(e3[2]) ;
    v3[0] = a2*st*st + b2*ct*ct ;
    v3[1] = a2*ct*ct + b2*st*st ;
    v3[2] = ct*st*(a2-b2) ;
    return(0) ;
}

int tr_vare3(v3, e3)
/*+++
.PURPOSE   Transformation of (covariance, variance) into error ellipse
.RETURNS   0
.REMARKS   e3[0] = a,      e3[1] = b,      e3[2] = theta ; 
	   v3[0] = var(x), v3[1] = var(y), v3[2] = cov(x,y)
.METHOD    If (x,y) have the input covariance matrix e3, (X,Y) defined by
		X = sin(theta)*x + cos(theta)*y
		Y =-cos(theta)*x + sin(theta)*y
	   have the covariances defined by (st and ct are sin and cos of theta):
	   var(X) = st^2^*var(x) + ct^2^*var(y) + 2.*ct*st*cov(x,y)
	   var(Y) = ct^2^*var(x) + st^2^*var(y) - 2.*ct*st*cov(x,y)
	   cov(X,Y) = ct*st*(var(y)-var(x)) + cov(x,y)*(s^2^-c^2^)
---*/
  double v3[2] ;	/* IN: Variances + cov	*/
  double e3[3] ;	/* OUT: Error ellipse	*/
{
  double theta, ct, st, a2, b2, tmp ;
    if ((v3[2]*v3[2]/(v3[0]*v3[1])) < 1.e-10) theta = 0; /* V3.44: circular */
    else if (v3[0] == v3[1]) theta = 45. ;
    else theta = 0.5*atand(2.*v3[2]/(v3[1] - v3[0])) ;
    ct = cosd(theta) ;   st = sind(theta) ;
    a2 = st*st*v3[0] + ct*ct*v3[1] + (tmp = 2.*ct*st*v3[2]) ;
    b2 = ct*ct*v3[0] + st*st*v3[1] - tmp ;
    if (a2>=b2) e3[0] = a2, e3[1] = b2, e3[2] = theta ;
    else        e3[0] = b2, e3[1] = a2, e3[2] = theta+90. ;
    e3[0] = sqrt(e3[0]) ;
    e3[1] = sqrt(e3[1]) ;
    if (e3[2] < 0) 	e3[2] += 180. ;
    if (e3[2] >= 180.) 	e3[2] -= 180. ;
    return(0) ;
}

static double *mp36(A, B, R)
/*+++
.PURPOSE   Rotate a 6-matrix from a 3-matrix
.RETURNS   Address of Result
.REMARKS   A transformed to 6x6 with  A6 = (A 0 , 0 A)
---*/
  double A[3][3] ; 	/* IN: Matrix #1	*/
  double B[6][6] ; 	/* IN: Matrix #2 	*/
  double R[6][6] ; 	/* OUT: Result Matrix 	*/
{
  int i,j,k ;
    for (i=0; i<3; i++) for (j=0; j<6; j++) 
    	R[i][j] = A[i][0]*B[0][j] + A[i][1]*B[1][j] + A[i][2]*B[2][j] ;
    for (i=3; i<6; i++) for (j=0; j<6; j++) 
    	k = i-3, 
    	R[i][j] = A[k][0]*B[3][j] + A[k][1]*B[4][j] + A[k][2]*B[5][j] ;
    return(&R[0][0]) ;
}

static double *mp361(A, B, R)
/*+++
.PURPOSE   Rotate a 6-matrix from a 3-matrix
.RETURNS   Address of Result
.REMARKS   A transformed to 6x6 with  A6 = (tA 0 , 0 tA)
---*/
  double A[3][3] ; 	/* IN: Matrix #1	*/
  double B[6][6] ; 	/* IN: Matrix #2 	*/
  double R[6][6] ; 	/* OUT: Result Matrix 	*/
{
  int i,j,k ;
    for (i=0; i<3; i++) for (j=0; j<6; j++) 
    	R[i][j] = A[0][i]*B[0][j] + A[1][i]*B[1][j] + A[2][i]*B[2][j] ;
    for (i=3; i<6; i++) for (j=0; j<6; j++) 
    	k = i-3, 
    	R[i][j] = A[0][k]*B[3][j] + A[1][k]*B[4][j] + A[2][k]*B[5][j] ;
    return(&R[0][0]) ;
}

static VOID RVtR(R, V, m)
/*+++
.PURPOSE   Computes the matrix product (R) . (V) . (tR)
.RETURNS   ---
.REMARKS   (i,j) element = SS(k,l) R_ik_ R_jl_ V_kl_
---*/
  double R[6][6] ; 	/* IN: Matrix #3	*/
  double V[6][6] ; 	/* IN: Matrix #2	*/
  double m[6][6] ; 	/* OUT: Result Matrix 	*/
{
  double t ; int i,j,k,l ;
    for (i=0; i<6; i++) for (j=0; j<6; j++) {
    	t = 0 ;
    	for (k=0; k<6; k++) for (l=0; l<6; l++) t += R[i][k]*R[j][l]*V[k][l] ;
    	m[i][j] = t ;
    }
}

static VOID tRVR(R, V, m)
/*+++
.PURPOSE   Computes the matrix product (tR) . (V) . (R)
.RETURNS   ---
.REMARKS   (i,j) element = SS(k,l) R_ki_ R_lj_ V_kl_
---*/
  double R[6][6] ; 	/* IN: Matrix #3	*/
  double V[6][6] ; 	/* IN: Matrix #2	*/
  double m[6][6] ; 	/* OUT: Result Matrix 	*/
{
  double t ; int i,j,k,l ;
    for (i=0; i<6; i++) for (j=0; j<6; j++) {
    	t = 0 ;
    	for (k=0; k<6; k++) for (l=0; l<6; l++) t += R[k][i]*R[l][j]*V[k][l] ;
    	m[i][j] = t ;
    }
}

/*============================================================================*
 *			Conversion of `s' structure
 *============================================================================*/
int tr_setframe(s, frame, prec_est, equinox, epoch)
/*+++
.PURPOSE   Initial "setup" of a spos structure, with Frame, Equinox,
		and default Precision / Estimated Values
.RETURNS   Output frame code.
.REMARKS   v[0] = rv2v0(Vr,plx), Vr in km/s and plx (parallax) in arcsec
---*/
  struct spos *s;	/* OUT: Position in Phase space		*/
  int frame ;		/* IN:  Frame type, e.g. _COO_FK5_	*/
  int prec_est ;	/* IN:  Combination precision+_COO_EST	*/
  double equinox;	/* IN:  Equinox (ECL / FK4 / FK5) 	*/
  double epoch;		/* IN:  Epoch to set.		 	*/
{
  char *e ;
    memset(s, 0, sizeof(struct spos)) ;
    switch(s->frame = frame) {
      case _COO_FK4_:
	s->equinox = equinox ? equinox : BYtoJY(1950.) ;
	break ;
      case _COO_FK5_:
      case _COO_ECL_:
	s->equinox = equinox ;
	break ;
      case _COO_GAL_:
      case _COO_SUPERGAL_:
      case _COO_ICRS_:
	break ;
      default:
	fprintf(stderr, "****tr_setframe indicates a bad (%d) frame\n",
	    frame) ;
    }
    s->epoch = s->epocho = epoch ;
    s->precision = prec_est & 0xf ;
    e = (char *)0 ;
    if (s->precision <= 0) e = "unknown position" ;
    if (s->precision > 10) e = "accuracy better than micro-arcsec" ;
    if (e) fprintf(stderr, "++++tr_setframe called with precision %d (%s)\n", 
	s->precision, e) ;
    if ((frame <= 0) || (frame > 6)) fprintf(stderr, 
	"****tr_setframe indicates a bad (%d) frame\n", frame) ;

    /* The existence of parameters is flagged. */
    s->flagset = (prec_est>>8)^0x37 ;

    return(s->frame) ;
}

int tr_qvs(q, v, s)
/*+++
.PURPOSE   Transformation of Position + Mouvement into spos structure
.RETURNS   Output frame code.
.REMARKS   v[0] = rv2v0(Vr,plx), Vr in km/s and plx (parallax) in arcsec
	v may be null (zero proper motion assumed)
---*/
  double q[2];	/* IN: Position ra + dec	*/
  double v[3];	/* IN: Velocity in Local Frame ("/yr)	*/
  struct spos *s;	/* OUT: Position in Phase space	*/
{
    tr_oR(q, Rl);			/* Rotation matrix for velocity	*/
    s->u[0] = Rl[0][0];
    s->u[1] = Rl[0][1];
    s->u[2] = Rl[0][2];
    if (v) {
	tr_uu1(v, s->ud, Rl);		/* Velocity in General Frame	*/
        s->ud[0] /= ARCSEC;			/* To radians	*/
        s->ud[1] /= ARCSEC;			/* To radians	*/
        s->ud[2] /= ARCSEC;			/* To radians	*/
    }
    else s->ud[0] = s->ud[1] = s->ud[2] = 0;
    return(s->frame);
}

int tr_2s (ep, q, e3q, p, e3p, plx, rv, s)
/*+++
.PURPOSE   Transformation of astrometric elements into spos structure.
.RETURNS   Output frame code (ICRS)
.REMARKS   Not initialized in s are: equinox / frame / precision
	   plx and rv may be NULL.
---*/
  double ep[2] ;	/* IN: Epoch of data + Epoch of Observation	*/
  double q[2] ; 	/* IN: RA Dec (degrees, fields H8, H9) at ep[0]	*/
  double e3q[3];	/* IN: Error ellipse on pos (mas, mas, deg)	*/
  double p[2] ;		/* IN: Proper motions in mas/yr			*/
  double e3p[3];	/* IN: Error ellipse on pm (mas/yr,mas/yr, deg)	*/
  double plx[2] ;	/* IN: Parallax + its mean error (mas)		*/
  double rv[2] ;	/* IN: Radial velocity + mean error (km/s)	*/
  struct spos *s;	/* OUT: Position + Variances in Phase space	*/
{
  double v[3], e3[3], var[6][6], t; int i, j ;
    s->epoch  = ep[0] ;
    s->epocho = ep[1] ;
    s->flags = s->flagset ;
    memset(s->cov, 0, sizeof(s->cov)) ;
    memset(s->ud,  0, sizeof(s->ud)) ;
    tr_oR(q, Rl);			/* Rotation matrix for velocity	*/
    s->u[0] = Rl[0][0];
    s->u[1] = Rl[0][1];
    s->u[2] = Rl[0][2];
    v[0] = v[1] = v[2] = s->plx = s->e_plx = 0 ;
    if (!rv)  rv  = def_rv,  s->flags &= ~2 ;
    if (!plx) plx = def_plx, s->flags &= ~1 ;

    s->plx   = plx[0] ;
    s->e_plx = plx[1] ;
    s->rv    = rv[0] ;
    s->e_rv  = rv[1] ;

    /* There is a problem when the parallax is too small 	*/
    v[0] = rv2v0(rv[0],s->plx);	/* 3rd component of pmotion	*/
    if ((plx[0] <= plx[1]) || (plx[0] <= zero_plx)) v[0] = 0 ;
    
    if (p) {
	/* s->flags |= 4 ; */
	v[1] = p[0] , v[2] = p[1] ;
        pr4("....Velocity mas/yr", "%10.3f", v, rv[0]) ;
        tr_uu1(v, s->ud, Rl);		/* Velocity in General Frame	*/
        s->ud[0] /= (1000.*ARCSEC);	/* To radians/yr..		*/
        s->ud[1] /= (1000.*ARCSEC);	
        s->ud[2] /= (1000.*ARCSEC);
    }
    else {
	s->flags &= ~4 ;
	if (s->frame == _COO_FK4_)	/* Compute Default FK4 PMotion	*/
	    fk4_s0(s) ;
    }

    /* Transform error ellipses into covariance matrices */
    /* Covariances for position */
    if (!e3p) s->flags &= ~0x20 ;
    if (!e3q) {
	s->flags &= ~0x10 ;
        if (!e3p) return(s->frame) ;
        e3q = e3 ;
	i = s->precision ;
    	if (i >= ITEMS(def_err)) i = ITEMS(def_err)-1 ;
    	e3q[0] = e3q[1] = def_err[i] ;
    	e3q[2] = 90.; 
    }
    /* else s->flags |= 0x10 ; */

    /* Prepare covariance matrices + Rotation in 6-D space */
    memset(var, 0, sizeof(var)) ;
    memset(R6,  0, sizeof(R6)) ;
    for (i=0; i<3; i++) for (j=0; j<3; j++) 
	R6[i+3][j+3] = R6[i][j] = Rl[i][j] ;

    /* Covariance of position part (rows 0 1 2 ) */
    tr_e3var(e3q, e3) ;
    var[1][1] = e3[0] ; var[2][2] = e3[1] ; 
    var[1][2] = var[2][1] = e3[2] ;

    /* Covariances for proper motions (expressed in mas/yr) */
    if (e3p) {
        /* s->flags |= 0x20 ; */
        tr_e3var(e3p, e3) ;		/* Ellipse to variance	*/
        var[3][3] = rv2v0(rv[1],plx[1]) ;  var[3][3] *= var[3][3] ;
        var[4][4] = e3[0] ; var[5][5] = e3[1] ; 
        var[4][5] = var[5][4] = e3[2] ;
    }

    /* If covariance given for epoch ep[1], propagate error, i.e. cov.
       | 1  0  0  t  0  0 |           | 1  0  0  0  0  0 |    
       | 0  1  0  0  t  0 |           | 0  1  0  0  0  0 |
       | 0  0  1  0  0  t | *  var  * | 0  0  1  0  0  0 |
       | 0  0  0  1  0  0 |           | t  0  0  1  0  0 |
       | 0  0  0  0  1  0 |           | 0  t  0  0  1  0 |
       | 0  0  0  0  0  1 |           | 0  0  t  0  0  1 |
    */

    if (ep[0] != ep[1]) {
	t = ep[0] - ep[1] ;
	for (j=0; j<3; j++) for (i=0; i<6; i++)
	    var[i][j] += t*var[i][j+3] ;
	for (i=0; i<3; i++) for (j=0; j<6; j++) 
	    var[i][j] += t*var[i+3][j] ;
    }

    /* Rotate the covariance matrix to fixed frame */
    tRVR(R6, var, s->cov) ;

    return(s->frame) ;
}

int icrs_2s (ep, q, e3q, p, e3p, plx, rv, s)
/*+++
.PURPOSE   Transformation of Hipparcos astrometric elements into spos structure.
.RETURNS   Output frame code (ICRS)
---*/
  double ep[2] ;	/* IN: Epoch of data + Epoch of Observation	*/
  double q[2] ; 	/* IN: RA Dec in degrees (fields H8, H9)	*/
  double e3q[3];	/* IN: Error ellipse on pos (mas, mas, deg)	*/
  double p[2] ;		/* IN: Proper motions in mas/yr			*/
  double e3p[3];	/* IN: Error ellipse on pm (mas/yr,mas/yr, deg)	*/
  double plx[2] ;	/* IN: Parallax + its mean error (mas)		*/
  double rv[2] ;	/* IN: Radial velocity + mean error (km/s)	*/
  struct spos *s;	/* OUT: Position + Variances in Phase space	*/
{
    s->frame = _COO_ICRS_ ;
    s->precision = 10 ;		/* Hipparcos gives 10^-5^ arcsec  */
    s->equinox = 2000. ;	/* Useless, but... */
    return(tr_2s(ep, q, e3q, p, e3p, plx, rv, s)) ;
}

int hip_2s (a, c, rv, s)
/*+++
.PURPOSE   Transformation of Hipparcos astrometric elements into spos structure.
.RETURNS   Output frame code (ICRS)
.REMARKS   a = RA, Dec (deg), plx, mua, mud (mas)
	   c = Correlations
---*/
  double a[5] ; 	/* IN: Hipp. Astrometric vector H8,9,11,12,13	*/
  double c[15];		/* IN: Sigmas + correlations , fields H14 - H28	*/
  double rv[2] ;	/* IN: Radial velocity + mean error (km/s)	*/
  struct spos *s;	/* OUT: Position + Variances in Phase space	*/
{
  double v[3], cov[6][6]; int i, j ;
    s->frame = _COO_ICRS_ ;
    s->precision = 10 ;		/* Hipparcos gives 10^-5^ arcsec  */
    s->equinox = 2000. ;	/* Useless, but... */
    s->epoch  = s->epocho = 1991.25 ;
    s->flags  = s->flagset = 5|0x30 ;
    s->unused = 0 ;
    memset(s->cov, 0, sizeof(s->cov)) ;
    tr_oR(a, Rl);			/* Rotation matrix for velocity	*/
    s->u[0] = Rl[0][0];
    s->u[1] = Rl[0][1];
    s->u[2] = Rl[0][2];
    if (rv) s->flags |= 2 ;
    else  rv = def_rv ;
    s->plx   = a[2] ;
    s->e_plx = c[2] ;
    s->rv    = rv[0] ;
    s->e_rv  = rv[1] ;
    v[0] = rv2v0(rv[0],s->plx);		/* 3rd component of pmotion	*/
    v[1] = a[3] ;
    v[2] = a[4] ;
    pr4("....Velocity mas/yr", "%10.3f", v, rv[0]) ;
    tr_uu1(v, s->ud, Rl);		/* Velocity in General Frame	*/
    s->ud[0] /= (1000.*ARCSEC);		/* To radians/yr..		*/
    s->ud[1] /= (1000.*ARCSEC);	
    s->ud[2] /= (1000.*ARCSEC);

    /* Prepare covariance matrices + Rotation in 6-D space */
    memset(cov, 0, sizeof(cov)) ;
    memset(R6,  0, sizeof(R6)) ;
    for (i=0; i<3; i++) for (j=0; j<3; j++) 
	R6[i+3][j+3]= R6[i][j] = Rl[i][j] ;
    
    cov[1][1] = c[0] ;					/* e_RAdeg */
    cov[2][2] = c[1] ;					/* e_DEdeg */
    cov[3][3] = rv2v0(rv[1],c[2]) ;			/* e_v0	   */
    cov[4][4] = c[3]; 					/* e_pma   */
    cov[5][5] = c[4]; 					/* e_pmd   */
    cov[1][2] = cov[2][1] = c[5] *cov[1][1]*cov[2][2]; 	/* DE:RA 	*/
    cov[1][4] = cov[4][1] = c[8] *cov[1][1]*cov[4][4];	/* pmRA:RA	*/
    cov[2][4] = cov[4][2] = c[9] *cov[2][2]*cov[4][4];	/* pmRA:DE	*/
    cov[1][5] = cov[5][1] = c[11]*cov[1][1]*cov[5][5];	/* pmDE:RA	*/
    cov[2][5] = cov[5][2] = c[12]*cov[2][2]*cov[5][5];	/* pmDE:DE	*/
    cov[4][5] = cov[5][4] = c[14]*cov[4][4]*cov[5][5];	/* pmDE:pmRA	*/

    cov[1][1] *= cov[1][1] ;
    cov[2][2] *= cov[2][2] ;
    cov[3][3] *= cov[3][3] ;
    cov[4][4] *= cov[4][4] ;
    cov[5][5] *= cov[5][5] ;

    /* Rotate the covariance matrix to fixed frame */
    tRVR(R6, cov, s->cov) ;

    return(s->frame) ;
}

int fk4_2s (ep, q, e3q, p, e3p, plx, rv, s)
/*+++
.PURPOSE   Transformation of FK4 astrometric elements into spos structure
.RETURNS   Output frame code (FK5)
.REMARKS   Not initialized in s are: / precision. Assumed J2000 equinox.
---*/
  double ep[2] ;	/* IN: Epoch of data + Epoch of Observation	*/
  double q[2] ; 	/* IN: RA Dec in degrees (fields H8, H9)	*/
  double e3q[3];	/* IN: Error ellipse on pos (mas, mas, deg)	*/
  double p[2] ;		/* IN: Proper motions in mas/yr			*/
  double e3p[3];	/* IN: Error ellipse on pm (mas/yr,mas/yr, deg)	*/
  double plx[2] ;	/* IN: Parallax + its mean error (mas)		*/
  double rv[2] ;	/* IN: Radial velocity + mean error (km/s)	*/
  struct spos *s;	/* OUT: Position + Variances in Phase space	*/
{
    s->frame = _COO_FK4_ ;
    s->equinox = BYtoJY(1950.) ;	
    if (s->precision > 6) s->precision = 6 ;
    return(tr_2s(ep, q, e3q, p, e3p, plx, rv, s)) ;
}

int fk5_2s (ep, q, e3q, p, e3p, plx, rv, s)
/*+++
.PURPOSE   Transformation of FK5 astrometric elements into spos structure
.RETURNS   Output frame code (FK5)
.REMARKS   Not initialized in s are: / precision. Assumed J2000 equinox.
---*/
  double ep[2] ;	/* IN: Epoch of data + Epoch of Observation	*/
  double q[2] ; 	/* IN: RA Dec in degrees (fields H8, H9)	*/
  double e3q[3];	/* IN: Error ellipse on pos (mas, mas, deg)	*/
  double p[2] ;		/* IN: Proper motions in mas/yr			*/
  double e3p[3];	/* IN: Error ellipse on pm (mas/yr,mas/yr, deg)	*/
  double plx[2] ;	/* IN: Parallax + its mean error (mas)		*/
  double rv[2] ;	/* IN: Radial velocity + mean error (km/s)	*/
  struct spos *s;	/* OUT: Position + Variances in Phase space	*/
{
    s->frame = _COO_FK5_ ;
    s->equinox = 2000 ;	
    return(tr_2s(ep, q, e3q, p, e3p, plx, rv, s)) ;
}

/*============================================================================*/
int tr_sqv(s, q, v)
/*+++
.PURPOSE   Transformation of spos structure (position in phase space)
		to Position + Velocity
.RETURNS   Frame in which Position + Velocity are expressed
	v may be null (zero proper motion assumed)
---*/
  struct spos *s;	/* IN: Position in Phase space	*/
  double q[2];	/* OUT: Position ra + dec	*/
  double v[3];	/* OUT: Velocity in Local Frame ("/yr)	*/
{
  int	i;
	
    tr_uR(s->u, Rl);		/* Rotation matrix for velocity	*/
    tr_uo(s->u, q);		/* Convert to Angles		*/
    if (v) {
        tr_uu(s->ud, v, Rl);	/* Velocity in Local Frame	*/
        for (i=0; i<3; i++)	v[i] *= ARCSEC;	/* Radians to "	*/
    }
    return(s->frame);
}

int tr_s2 (s, ep, q, e3q, p, e3p, plx, rv)
/*+++
.PURPOSE   Transformation spos structure into astrometric elements 
.RETURNS   Output frame code 
.REMARKS   Not initialized in s are: equinox / frame / precision
---*/
  struct spos *s;	/* IN: Position + Variances in Phase space	*/
  double ep[2] ;	/* OUT: Epoch of data + Epoch of Observation	*/
  double q[2] ; 	/* OUT: RA Dec in degrees (fields H8, H9)	*/
  double e3q[3];	/* OUT: Error ellipse on pos (mas, mas, deg)	*/
  double p[2] ;		/* OUT: Proper motions in mas/yr		*/
  double e3p[3];	/* OUT: Error ellipse on pm (mas/yr,mas/yr, deg)*/
  double plx[2] ;	/* OUT: Parallax + its mean error (mas)		*/
  double rv[2] ;	/* OUT: Radial velocity + mean error (km/s)	*/
{
  double v[3], v0, var[6][6], pm2 ; int i, j ;
  
    if (ep) {
        ep[0] = s->epoch  ;
        ep[1] = s->flags&0x10 ? s->epocho : s->epoch ;
    }

    /* Coordinates & Proper Motions */
    tr_uR(s->u, Rl);		/* Rotation matrix for velocity	*/
    tr_uo(s->u, q);		/* Convert to Angles		*/
    tr_uu(s->ud, v, Rl);	/* Velocity in Local Frame	*/
    v[0] *= (1000.*ARCSEC) ;
    v[1] *= (1000.*ARCSEC) ;
    v[2] *= (1000.*ARCSEC) ;
    pr4("....Velocity mas/yr", "%10.3f", v, rv[0]) ;
    v0 = v[0] ;			/* Keep to save rv in output	*/
    pm2 = v[1]*v[1] + v[2]*v[2];	/* Square of total pm	*/
    if (p) {
        p[0] = v[1] ;
        p[1] = v[2] ;
    }
    
    /* Error Ellipses */
    memset(R6,  0, sizeof(R6)) ;
    if (e3q || e3p) {
        for (i=0; i<3; i++) for (j=0; j<3; j++) 
	    R6[i+3][j+3]= R6[i][j] = Rl[i][j] ;
	RVtR(R6, s->cov, var) ;
    }

    if (e3q) {
    	if (s->flags&0x10) {	/* Rotate cov matrix to local	*/
    	    v[0] = var[1][1] ;
    	    v[1] = var[2][2] ;
    	    v[2] = var[1][2] ;
    	    tr_vare3(v, e3q) ;
    	}
    	else {
	    i = s->precision ;
    	    if (i >= ITEMS(def_err)) i = ITEMS(def_err)-1 ;
    	    e3q[0] = e3q[1] = def_err[i] ;
    	    e3q[2] = 90.; 
    	}
    }
    if (e3p) {
    	if (s->flags&0x20) {
    	    v[0] = var[4][4] ;
    	    v[1] = var[5][5] ;
    	    v[2] = var[4][5] ;
    	    tr_vare3(v, e3p) ;
    	}
    	else {	/* Defaut error: 5%, ot 30mas/yr */
	    e3p[0] = e3p[1] = 0.05*sqrt(pm2);
	    if (e3p[0] < 30) e3p[0] = 30;
	    if (e3p[1] < 30) e3p[1] = 30;
	    e3p[2] = 90 ;
	}
    }
    
    /* Parallax */
    if (plx) {
    	if (s->flags&1) plx[0] = s->plx, plx[1] = s->e_plx ;
    	else plx[0] = plx[1] = 0 ;
    }

    /* Radial Velocity */
    if (rv) {
	rv[0] = rv[1] = 0 ;
	if (s->flags&2) {
	    rv[0] = s->rv ;
	    if ((s->flags&1) && (s->plx >= zero_plx) && (s->plx > s->e_plx)) 
		rv[0] = v0 / s->plx / kms_pc ;
	}
    	if ((s->flags&0x20) && (s->plx > s->e_plx))
    	     rv[1] = sqrt(var[3][3]) / s->e_plx / kms_pc ;
        else if (s->flags&2) rv[1] = s->e_rv ;
    }


    /* The returned status must set the _COO_EST_xxx flags for estimated
       values; take care that the 0x37 value has to be changed when
       new parameters will be part of the spos structure
    */
       
    return(s->frame|((s->flags^0x37)<<8)) ;
}

/*============================================================================*
 *			E-term computation
 *============================================================================*/

static VOID eterm(y, ev)
/*+++++++++++
.PURPOSE   Computation of e-term at specified date
.RETURNS   -
.METHOD		Reference: Explanatory Suppl. to the Astron. Ephemeris, 
	section 2D, p. 48;	P.T. Wallace, Starlink.
------------*/
  double y; 	/* IN: Epoch, in Julian Years	*/
  double ev[3]; 	/* OUT: E-term modification	*/
{
  double t, e, L, eps, cL;
  	
    t   = (y - 1900.e0)/1.e2;		/* 	Centuries	*/
    eps = 23.452   - 0.013*t;		/* Mean obliquity	*/
    e   = 0.016751 - 0.000042*t;	/* Eccentricity		*/
    L   = 281.221  + 1.719*t;		/* Mean longitude of perihelion	*/  
  
    e  *= (20.496/ARCSEC);		/* Component of aberration	*/
    cL  = cosd(L);
    ev[0] =  e * sind(L);
    ev[1] = -e * cL * cosd(eps);
    ev[2] = -e * cL * sind(eps);

#ifdef DEBUG
    printf("eterm: %12.8f %12.8f %12.8f\n", ev[0], ev[1], ev[2]);
#endif
}

/*============================================================================*
 *			Math utilities
 *============================================================================*/

static double add(u,v,t)
/*+++
.PURPOSE Add vector t*v to unit vector; vectors are rescaled.
.RETURNS The norm of (u+vt), which scales all vectors.
.REMARKS NO ADDITION WHEN VECTOR == 0!
---*/
  double u[3];	/* MOD: Unit vector	*/
  double v[3];	/* MOD: Vector to add	*/
  double t;	/* IN: Time interval	*/
{
  double  f;
  int	i;

#ifdef DEBUG
    printf("add: u=%12.8f %12.8f %12.8f (in)\n", u[0], u[1], u[2]);
    printf("     v=%12.8f %12.8f %12.8f (in)  t=%12.5f\n", v[0], v[1], v[2], t);
#endif
    if ((u[0] == 0) && (u[1] == 0) && (u[1] == 0))	/* V3.42 */
	return(0.);
    for (i=0; i<3; i++)
    	u[i] += t*v[i];

    f = hypot(hypot(u[0], u[1]), u[2]);
  
    for (i=0; i<3; i++)	u[i] /= f;
    for (i=0; i<3; i++)	v[i] /= f; 
  
#ifdef DEBUG
    printf("     u=%12.8f %12.8f %12.8f (out) f=%12.9f\n", u[0], u[1], u[2], f);
    printf("     v=%12.8f %12.8f %12.8f (out) t=%12.5f\n", v[0], v[1], v[2], t);
#endif
    return(f) ;
}

/*============================================================================*/
static VOID sub(u,v,t)
/*+++
.PURPOSE Substract vector t*v to unit vector
---*/
  double u[3];	/* MOD: Unit vector	    */
  double v[3];	/* MOD: Vector to substract */
  double t;	/* IN: Time interval	    */
{
  double  f;
  int	i;

#ifdef DEBUG
    printf("sub: u=%12.8f %12.8f %12.8f (in)\n", u[0], u[1], u[2]);
    printf("     v=%12.8f %12.8f %12.8f (in)  t=%12.5f\n", v[0], v[1], v[2], t);
#endif
    for (i=0; i<3; i++)
    	u[i] -= t*v[i];

    f = hypot(hypot(u[0], u[1]), u[2]);
    for (i=0; i<3; i++)	u[i] /= f;
    for (i=0; i<3; i++)	v[i] /= f; 
  
#ifdef DEBUG
    printf("     u=%12.8f %12.8f %12.8f (out) f=%12.9f\n", u[0], u[1], u[2], f);
    printf("     v=%12.8f %12.8f %12.8f (out) t=%12.5f\n", v[0], v[1], v[2], t);
#endif
}

/*============================================================================*/
static VOID rotate(s, R, inv)
/*+++
.PURPOSE Rotate unit vector + derivative
---*/
  struct spos *s;	/* MOD: Pos + derivative	*/
  double R[3][3];	/* IN: Rotation Matrix		*/
  int    inv;		/* IN: 1 to take inverse matrix	*/
{
  static double var[6][6] ;
  int  i, j ;

    if (s->flags&0x30){	/* Rotate also the error matrix	*/
    	memcpy(var, s->cov, sizeof(var)) ;
    	memset(R6,  0, sizeof(R6)) ;
	for (i=0; i<3; i++) for (j=0; j<3; j++) 
	    R6[i+3][j+3] = R6[i][j] = R[i][j] ;
    }
    
    if (inv) {		/* Transformation  y = tR . x	*/
  	tr_uu1(s->u,  s->u,  R),
  	tr_uu1(s->ud, s->ud, R);
  	if (s->flags&0x30) tRVR(R6, var, s->cov) ;
    }
    else {		/* Transformation  y = R . x	*/
  	tr_uu(s->u,  s->u,  R),
  	tr_uu(s->ud, s->ud, R);
  	if (s->flags&0x30) RVtR(R6, var, s->cov) ;
    }
}

#ifdef TEST		/* Testing ICRS Transformations	*/
int tr_ss(struct spos *s0, struct spos *s1)
{
  double  ev[3], f, t ;
  static double T[6][6] ;
  int	i;

    for (i=0; i<6; i++)	s1->u[i] = s0->u[i];
    s1->precision = s0->precision;
    s1->flags     = s0->flags ;
    t = s1->epoch - s0->epoch ;

    f = add(s1->u, s1->ud, t) ;
    if (s1->flags) {			/* Parallax exists 	*/
  	s1->equinox= s0->equinox;
  	s1->epocho = s0->epocho ;
  	s1->plx    = s0->plx   / f ;
	s1->e_plx  = s0->e_plx / f ;
  	s1->rv     = s0->rv ;
	s1->e_rv   = s0->e_rv ;
    }
    if (s1->flags&0x30) {
        memset(T, 0, sizeof(T)) ;
        T[0][0] = T[1][1] = T[2][2] = 
        T[3][3] = T[4][4] = T[5][5] = 1./f ;
        T[0][3] = T[1][4] = T[2][5] =  t/f ;
	RVtR(T, s0->cov, s1->cov) ;
    }

    return(s1->frame);
}

#else
/*============================================================================*
 *			General Coordinate Transformation
 *============================================================================*/
int tr_ss(s0, s1)
/*+++
.PURPOSE General Coordinate Transformation from frame s0->frame to frame 
	s1->frame.
.METHOD  For same frame, use standard transformations;
	use always J2000 intermediate frame for other transformations.
.RETURNS The frame in which s1 is expressed. A value of 0 is a bad frame.
.REMARKS If input = ICRS, keep it !
---*/
  struct spos *s0;/* IN: Vector in Phase Space (input)	*/
  struct spos *s1;/* OUT: Vector in Phase Space (output)	*/
{
  double  ev[3], f, t, *pR, *pT, *pv;
  double T[6][6], R[6][6] ;
  static  struct spos s ;
  int	i;

    t = s1->epoch - s0->epoch ;		/* Change of Epoch...		*/
    s1->precision = s0->precision;
    s1->flags     = s0->flags ;
    pT = pR = (double *)0 ;
    if (s1->flags&0x30) {		/* Prepare error propagation...	*/
        memset(T, 0, sizeof(T)) ;
        T[0][0] = T[1][1] = T[2][2] = 
        T[3][3] = T[4][4] = T[5][5] = 1 ;
        T[0][3] = T[1][4] = T[2][5] = t ;
    	pT = &T[0][0] ; pR = &R[0][0] ;
    }

    if (s0->frame == s1->frame) {
  					/* Add Proper Motion		*/
  	for (i=0; i<6; i++) s1->u[i] = s0->u[i] ;
  	f = add(s1->u, s1->ud, t) ;

  	switch(s0->frame) {
  	  case _COO_FK4_:		/* FK4 system (equatorial)	*/
		if ((y40 != s0->equinox) || (y41 != s1->equinox))
			y40 = s0->equinox, 
			y41 = s1->equinox, 
			preb_R(R44, JYtoBY(y40), JYtoBY(y41));
		eterm(s0->equinox, ev);		/* E-term to substract	*/
		sub(s1->u, ev, 1.e0);
		rotate(s1, R44, 0);
		if (pT)  			/* Propagate Errors	*/
		    pT = mp36(R44, pv=pT, pR), pR=pv ;
		eterm(s1->equinox, ev);		/* E-term to add	*/
		add(s1->u, ev, 1.e0);
		break;

	  case _COO_ICRS_:		/* ICRS Hipparcos, barycentric	*/
	  case _COO_GAL_: case _COO_SUPERGAL_:
	  	s1->equinox = s0->equinox ;
		break;

	  case _COO_FK5_:		/* FK5 system (equatorial)	*/
		if ((y50 != s0->equinox) || (y51 != s1->equinox))
			y50 = s0->equinox, 
			y51 = s1->equinox, 
			prej_R(R55, y50, y51);
		rotate(s1, R55, 0);
		if (pT)  			/* Propagate Errors	*/
		    pT = mp36(R55, pv=pT, pR), pR=pv ;
		break;

	  case _COO_ECL_:
	  	goto via_J2000;

	  default:	return(0);
  	}
  	goto FIN;
    }

	/* Convert first to J2000, epoch J2000 */

  via_J2000:

#ifdef DEBUG
    printf(".... Via J2000 ...\n");
#endif
  
    for (i=0; i<6; i++) s.u[i] = s0->u[i] ;
    s.flags = 0 ;
    s.precision = s0->precision ;
    
    switch(s0->frame) {
    case _COO_FK4_:			/* FK4 system (equatorial)	*/
	f = add(s.u, s.ud, 1950.e0 - JYtoBY(s0->epoch));
					/* Substract E-term		*/
	eterm(s0->equinox, ev); sub(s.u, ev, 1.e0);
	if (yB0 != s0->equinox)
		yB0 = s0->equinox, 
		preb_R(RB0, JYtoBY(yB0), 1950.e0);
	rotate(&s, RB0, 0);
	if (pT)  				/* Propagate Errors	*/
	    pT = mp36(RB0, pv=pT, pR), pR=pv ;
					/* Add E-term			*/
	eterm(BYtoJY(1950.e0), ev); add(s.u, ev, 1.e0);
	fk4_5s(&s, &s);			/* Convert to J2000 ep+eq	*/
	t = 2000. - BYtoJY(1950.) ;
	for (i=0; i<3; i++) ev[i] = s.u[i] + t*s.ud[i] ;
	f *= sqrt(scalar(ev, ev)) ;
	if (pT) {
	    if (!R45[0][0]) preb_R(R45, 1950., JYtoBY(2000.)) ;
	    pT = mp36(R45, pv=pT, pR), pR=pv ;
	}
	break;

    case _COO_ICRS_:			/* ICRS (Hipparcos) Frame	*/
	f = add(s.u, s.ud, 2000.e0 - s0->epoch);
	break ;

    case _COO_FK5_:			/* FK5 system (equatorial)	*/
	f = add(s.u, s.ud, 2000.e0 - s0->epoch);
	if (yJ0 != s0->equinox)
		yJ0 = s0->equinox, 
		prej_R(RJ0, yJ0, 2000.e0);
	rotate(&s, RJ0, 0);
	if (pT)  				/* Propagate Errors	*/
	    pT = mp36(RJ0, pv=pT, pR), pR=pv ;
	break;

    case _COO_GAL_:			/* Galactic coordinates		*/
	f = add(s.u, s.ud, 2000.e0 - s0->epoch);
	rotate(&s, gal_2000, 1);
	if (pT)  				/* Propagate Errors	*/
	    pT = mp361(gal_2000, pv=pT, pR), pR=pv ;
	break;
    	
    case _COO_SUPERGAL_:		/* SuperGalactic coordinates	*/
	if (GAL_2000[0][1] == 0.e0)		/* Compute GAL_2000	*/
	    tr_RR(gal_2000, GAL_2000, supergal);
	f = add(s.u, s.ud, 2000.e0 - s0->epoch);
	rotate(&s, GAL_2000, 1);
#ifdef DEBUG
	{ int i,j;
	for (i=0; i<3; i++) for (j=i; j<3; j++) 
	    printf("SUPERGal: i=%d, j=%d, scalar=%e\n", i, j, scalar(GAL_2000[i],GAL_2000[j])) ;
	}
#endif
	if (pT)  				/* Propagate Errors	*/
	    pT = mp361(GAL_2000, pv=pT, pR), pR=pv ;
	break;

    case _COO_ECL_:			/* Ecliptic coordinates		*/
	f = add(s.u, s.ud, 2000.e0 - s0->epoch);	
	if (yE0 != s0->equinox)
		yE0 = s0->equinox, 
		ecl_R(RE0, s0->equinox);
	rotate(&s, RE0, 1);		/* Equatorial, equinox yE0 	*/
	if (pT)  				/* Propagate Errors	*/
	    pT = mp361(RE0, pv=pT, pR), pR=pv ;
	if (yJ0 != s0->equinox)
		yJ0 = s0->equinox, 
		prej_R(RJ0, s0->equinox, 2000.e0);
	rotate(&s, RJ0, 0);
	if (pT)  				/* Propagate Errors	*/
	    pT = mp36(RJ0, pv=pT, pR), pR=pv ;
	break;
	
    default:	return(0);
    }

    s.equinox = s.epoch = 2000.; 
    s.frame = _COO_FK5_ ;

    /* Convert from (J2000, J2000)	*/

    switch(s1->frame) {
    case _COO_FK4_:			/* FK4 system (equatorial)	*/
	fk5_4s(&s, &s);			/* Convert to B1950 ep+eq	*/
	t = BYtoJY(1950.) - 2000. ;
	for (i=0; i<3; i++) ev[i] = s.u[i] + t*s.ud[i] ;
	f *= sqrt(scalar(ev, ev)) ;
	if (pT) {
	    if (!R54[0][0]) preb_R(R54, 1950., JYtoBY(2000.)) ;
	    pT = mp36(R54, pv=pT, pR), pR=pv ;
	}
	f *= add(s.u, s.ud, JYtoBY(s1->epoch) - 1950.e0);    /* Add PM	*/
	eterm(BYtoJY(1950.e0), ev);		/* E-term to remove	*/
	sub(s.u, ev, 1.e0);
	if (yB1 != s1->equinox)
		yB1 = s1->equinox, 
		preb_R(RB1, 1950.e0, JYtoBY(yB1));
	rotate(&s, RB1, 0);
	if (pT)  				/* Propagate Errors	*/
	    pT = mp36(RB1, pv=pT, pR), pR=pv ;
	eterm(s1->equinox, ev);			/* E-term to add	*/
	add(s.u, ev, 1.e0);
	break;

    case _COO_ICRS_:			/* FK5 system (equatorial)	*/
	f *= add(s.u, s.ud, s1->epoch - 2000.e0);	/* Add PM	*/
	s1->equinox = 2000.; 
	break;

    case _COO_FK5_:			/* FK5 system (equatorial)	*/
	f *= add(s.u, s.ud, s1->epoch - 2000.e0);	/* Add PM	*/
	if (yJ1 != s1->equinox)
		yJ1 = s1->equinox, 
		prej_R(RJ1, 2000.e0, yJ1);
	rotate(&s, RJ1, 0);
	if (pT)  				/* Propagate Errors	*/
	    pT = mp36(RJ1, pv=pT, pR), pR=pv ;
	break;

    case _COO_GAL_:			/* Galactic coordinates		*/
	f *= add(s.u, s.ud, s1->epoch - 2000.e0);	/* Add PM	*/
	rotate(&s, gal_2000, 0);
	if (pT)  				/* Propagate Errors	*/
	    pT = mp36(gal_2000, pv=pT, pR), pR=pv ;
	break;
    	
    case _COO_SUPERGAL_:		/* SuperGalactic coordinates	*/
	if (GAL_2000[0][1] == 0.e0)	
	    tr_RR(gal_2000, GAL_2000, supergal);
	f *= add(s.u, s.ud, s1->epoch - 2000.e0);	/* Add PM	*/
	rotate(&s, GAL_2000, 0);
	if (pT)  				/* Propagate Errors	*/
	    pT = mp36(GAL_2000, pv=pT, pR), pR=pv ;
	break;

    case _COO_ECL_:			/* Ecliptic coordinates		*/
	f *= add(s.u, s.ud, s1->epoch - 2000.e0);	/* Add PM	*/
	if (yJ1 != s1->equinox)
		yJ1 = s1->equinox, 
		prej_R(RJ1, 2000.e0, yJ1);
	rotate(&s, RJ1, 0);
	if (pT)  				/* Propagate Errors	*/
	    pT = mp36(RJ1, pv=pT, pR), pR=pv ;
	if (yE1 != s1->equinox)
		yE1 = s1->equinox, 
		ecl_R(RE1, yE1);
	rotate(&s, RE1, 0);	/* Ecliptic	*/
	if (pT)  				/* Propagate Errors	*/
	    pT = mp36(RE1, pv=pT, pR), pR=pv ;
	break;
	
    default:	return(0);
    }
    for (i=0; i<6; i++) s1->u[i] = s.u[i] ;

  FIN:

    /* .... Propagate parallax ... */
    /* V3.42: f==0 when position unknown ! */
    if (f < 0.1) f = 1.;
    if (s1->flags) {			/* Parallax exists 	*/
  	s1->epocho = s0->epocho ;
  	s1->plx    = s0->plx   / f ;
	s1->e_plx  = s0->e_plx / f ;
  	s1->rv     = s0->rv ;
	s1->e_rv   = s0->e_rv ;
    }
    /* .... Propagate Errors ....  */
    if (s1->flags&0x30) {
        for (i=0; i<6*6; i++) pT[i] /= f ;
	RVtR(pT, s0->cov, s1->cov) ;
    }

    return(s1->frame);
}

int tr_s(s, frame, equinox, epoch)
/*+++
.PURPOSE General Coordinate Transformation 
.RETURNS frame. A value of 0 is a bad frame.
.REMARKS equinox is ignored for GAL SUPERGAL, etc
---*/
  struct spos *s;	/* MOD: Vector in Phase Space 	*/
  int frame ;		/* IN: New frame (0=same)	*/
  double equinox ;	/* IN: New equinox		*/
  double epoch ;	/* IN: New epoch		*/
{
  struct spos s1 ; int r, len ;
    s1.frame = frame ? frame : s->frame ;
    s1.equinox = equinox ;
    s1.epoch   = epoch ;
    r = tr_ss(s, &s1) ;
    /* Copy Back to s */
    if (r <= 0) return(r) ;
    len = sizeof(s1) ;
    if (s->flags == 0) len -= 21*sizeof(double) ;
    memcpy(s, &s1, len) ;
    return(r) ;
}
#endif

/*============================================================================*
 *		Interpretation of Important Catalogues
 *============================================================================*/
#define byte(pos)	(rec-1+pos)

struct spos *hip_s(char *rec, double rv[2])
/*++++++++++++++++
.PURPOSE  Get next HIP record (Catalog I/239/hip_main)
.RETURNS  Structure / NULL
.REMARKS  
-----------------*/
{
  static struct spos s ;
  double atof(), a[5], c[15] ;

    memset(&s, 0, sizeof(s)) ;
    s.precision = 10 ;			/* 0.01mas    */

    /* Positions */
    a[0] = atof(byte(52)) ;	/* RA  */
    a[1] = atof(byte(65)) ;	/* Dec */
    a[2] = atof(byte(80)) ;	/* plx */
    a[3] = atof(byte(88)) ;	/* mua */
    a[4] = atof(byte(97)) ;	/* mud */

    /* Variances + Coorelations */
    c[0]  = atof(byte(106)) ;
    c[1]  = atof(byte(113)) ;
    c[2]  = atof(byte(120)) ;
    c[3]  = atof(byte(127)) ;
    c[4]  = atof(byte(134)) ;
    c[5]  = atof(byte(141)) ;
    c[6]  = atof(byte(147)) ;
    c[7]  = atof(byte(153)) ;
    c[8]  = atof(byte(159)) ;
    c[9]  = atof(byte(165)) ;
    c[10] = atof(byte(171)) ;
    c[11] = atof(byte(177)) ;
    c[12] = atof(byte(183)) ;
    c[13] = atof(byte(189)) ;
    c[14] = atof(byte(195)) ;

    hip_2s(a, c, rv, &s) ;
    return(&s) ;
}

struct spos *ppm_s(char *rec, double plx[2], double rv[2])
/*++++++++++++++++
.PURPOSE  Interpret PPM record (catalogues I/146, I/193, I/206, I/208)
.RETURNS  Structure / NULL
.REMARKS  
-----------------*/
{
  static struct spos s ;
  double atof();
  double ep[2], q[2], e3q[3], p[2], e3p[3] ;

    memset(&s, 0, sizeof(s)) ;
    s.precision = 7 ;			/* 0.01arcsec */

    /* Positions */
    q[0] = 15*(atof(byte(28)) + atof(byte(31))/60. + atof(byte(34))/3600.) ;
    q[1] =     atof(byte(43)) + atof(byte(46))/60. + atof(byte(49))/3600. ;
    if (*(byte(42)) == '-') q[1] = -q[1] ;
    p[0] = atof(byte(56))*cosd(q[1])*15000. ;	/* mua */
    p[1] = atof(byte(64))*1000. ;		/* mud */

    /* Error Ellipses */
    e3q[0] = atof(byte(74)) * 10.; 
    e3q[1] = atof(byte(77)) * 10.; 
    e3q[2] = 90. ;
    e3p[0] = atof(byte(80)) ; 
    e3p[1] = atof(byte(85)) ; 
    e3p[2] = 90. ;

    /* Epochs: of Position, of best Sigma on position */
    ep[0] = 2000. ;
    ep[1] = 1900. + (atof(byte(90)) + atof(byte(96)))/2. ;
    fk5_2s(ep, q, e3q, p, e3p, plx, rv, &s) ;
    return(&s) ;
}


#ifdef TEST
typedef struct spos *SPTR ;

static int a2v(char *s, double *v, int nv)
/*++++++++++++++++
.PURPOSE  Convert ascii to vector of double's
.RETURNS  Number of values (nv)
.REMARKS  Blanks(s) must separate numbers
-----------------*/
{
  char *p ; double *pv, atof() ;
    for (p=s, pv=v ; --nv >= 0 ; pv++) {
    	while (isspace(*p)) p++ ;
    	*pv = atof(p) ;
    	while (isgraph(*p)) p++ ;
    }
    return(pv-v) ;
}

#include <stdio.h>
main()
{
  SPTR (*get_rec)() ;
  FILE *f ;
  struct spos s1, s2, *s ;
  double ep[2], ep2, q[2], e3q[3], p[2], e3p[3], plx[2], rv[2] ;
  char buf[BUFSIZ] ; double v[3], atof() ;
  int i, n, est ;

  /* File with Hipparcos data ? */
  get_rec = 0 ;
  f = (FILE *)0 ;
  s = (struct spos *)0 ;
  rv[0] = rv[1] = 0. ;
  ep[0] = ep[1] = ep2 = 1991.25 ;
  printf(" 2. Epoch? "); gets(buf) ; if (buf[0]) a2v(buf,  &ep2, 1) ;
  while(1) {
    if (f) {
        if (!fgets(buf, sizeof(buf), f)) {
	    fclose(f);
	    f = (FILE *)0 ;
	    get_rec = 0 ;
	    s = (struct spos *)0 ;
	    continue ;
	}
	rv[0] = fabs(rv[0]) + 1. ;
	if (n&1) rv[0] = -rv[0] ;
	rv[1] = rv[0] * 0.1 ;
	if (rv[1] < 5) rv[1] = 5. ;
	n++ ;
	s = (*get_rec)(buf, rv) ;
	memcpy(&s1, s, sizeof(struct spos)) ;
	printf("....Next Hipparcos Record read (hit return) ") ;
	getchar() ;
	goto Show_Pos ;
    }
    printf("   (Hipparcos data possible with <hip.dat or ppm*.dat)\n") ;
    printf("   RA Dec(deg)? "); gets(buf) ; 
    if (buf[0] == '<') { 
        for (i=1; isspace(buf[i]); i++) ;
	f = fopen(buf+i, "r") ;
	if (!f) fprintf(stderr, "****"), perror(buf) ;
	if (strstr(buf, "hip")) get_rec = hip_s ;
	else if (strstr(buf, "ppm")) get_rec = ppm_s ;
	else fprintf(stderr, 
	    "****Can't understand which record type, assume hipparcos\n"),
	    get_rec = hip_s ;
	continue ;
    }
    if (buf[0]) a2v(buf, q, 2) ;
    est = 0 ;
    printf("   ErrEll(mas)? "); gets(buf) ; a2v(buf, e3q, 3) ;
				if (!buf[0]) est |= _COO_EST_me_pos_ ;
    printf("       pm(mas)? "); gets(buf) ; a2v(buf, p, 2) ;
				if (!buf[0]) est |= _COO_EST_pm_ ;
    printf("   ErrEll(mas)? "); gets(buf) ; a2v(buf, e3p, 3) ;
				if (!buf[0]) est |= _COO_EST_me_pm_ ;
    printf("   plx,_e(mas)? "); gets(buf) ; a2v(buf, plx, 2) ;
				if (!buf[0]) est |= _COO_EST_plx_ ;
    printf("   rv,_e(km/s)? "); gets(buf) ; a2v(buf,  rv, 2) ;
				if (!buf[0]) est |= _COO_EST_rv_ ;
    printf("----%12.8f%12.8f (%10.8f %10.8f %7.3f) plx: %6.3f %6.3f\n", 
    	q[0], q[1], e3q[0], e3q[1], e3q[2], plx[0], plx[1]) ;
    printf("    %12.6f%12.6f (%10.6f %10.6f %7.3f)  RV: %6.3f %6.3f\n", 
    	p[0], p[1], e3p[0], e3p[1], e3p[2], rv[0], rv[1]) ;
    icrs_2s(ep, q, e3q, p, e3p, plx, rv, &s1) ;
  Show_Pos:
    i = tr_s2(&s1, ep, q, e3q, p, e3p, plx, rv) ;
    printf("========Equinox=%.3f, Epoch=%.3f, osbEpoch=%.3f (flags=%04X)\n",
    	s1.equinox, s1.epoch, s1.epocho, i) ;
    printf("----%12.8f%12.8f (%10.3f %10.3f %7.3f) plx: %6.3f %6.3f\n", 
    	q[0], q[1], e3q[0], e3q[1], e3q[2], plx[0], plx[1]) ;
    printf("    %12.6f%12.6f (%10.5f %10.5f %7.3f)  RV: %6.3f %6.3f\n", 
    	p[0], p[1], e3p[0], e3p[1], e3p[2], rv[0], rv[1]) ;
    tr_setframe(&s2, s1.frame, 0, s1.equinox, ep2) ;
    /* s2.frame = s1.frame ;
       s2.epoch = ep2 ;
       s2.equinox = s1.equinox ;
    */
    tr_ss(&s1, &s2) ;
    i = tr_s2(&s2, ep, q, e3q, p, e3p, plx, rv) ;
    printf("========Equinox=%.3f, Epoch=%.3f, osbEpoch=%.3f (flags=%04X)\n",
    	s2.equinox, s2.epoch, s2.epocho, i) ;
    printf("----%12.8f%12.8f (%10.3f %10.3f %7.3f) plx: %6.3f %6.3f\n", 
    	q[0], q[1], e3q[0], e3q[1], e3q[2], plx[0], plx[1]) ;
    printf("    %12.6f%12.6f (%10.5f %10.5f %7.3f)  RV: %6.3f %6.3f\n", 
    	p[0], p[1], e3p[0], e3p[1], e3p[2], rv[0], rv[1]) ;
    tr_ss(&s2, &s1) ;
    i = tr_s2(&s1, ep, q, e3q, p, e3p, plx, rv) ;
    printf("=(Back)=Equinox=%.3f, Epoch=%.3f, osbEpoch=%.3f (flags=%06X)\n",
    	s1.equinox, s1.epoch, s1.epocho, i) ;
    printf("----%12.8f%12.8f (%10.3f %10.3f %7.3f) plx: %6.3f %6.3f\n", 
    	q[0], q[1], e3q[0], e3q[1], e3q[2], plx[0], plx[1]) ;
    printf("    %12.6f%12.6f (%10.5f %10.5f %7.3f)  RV: %6.3f %6.3f\n", 
    	p[0], p[1], e3p[0], e3p[1], e3p[2], rv[0], rv[1]) ;
  }
}
#endif
