/*+++++++++++++++       
.TYPE           Module
.IDENTIFICATION fk4fk5.c
.VERSION 1.0	05-Mar-1987: Creation .
.VERSION 1.1	26-Oct-1988: Use `s' structure
.VERSION 2.0	22-Feb-1991: Tested on PCs with TuboC
.VERSION 2.1	12-Jul-1998: Update all elements of 's' structure
.VERSION 2.2	28-Aug-2010: reentrent
.LANGUAGE       C
.AUTHOR         P.T Wallace [Starlink], Francois Ochsenbein [ESO-IPG]  
.KEYWORDS       Precession of Coordinates with proper motions,

.ENV            

.COMMENTS
This module contains routines to compute the
transformation from B1950 (FK4) to J2000 (FK5)
epochs and vice-versa. The modification concerns the position as well as the
velocity vector (radial velocity and proper motions).

\begin{TeX}

The velocity vector $\vec{v}(t)$ contains the velocity expressed in 
$''$/year as:

$$\begin{array}{ll}
  v_0 = k \pi V_{r0} 		& \\
  v_1 = \mu_{\alpha0} \cos \delta_0 &  	\\
  v_2 = \mu_{\delta0} 		&  	
\end{array}$$

The value of $k$  ($0.21094502$) when $\pi$ is
expressed in " and $V_r$ in km/s
is defined in {\tt ASTRO.H} as the mnemonic {\tt km\_s\_pc}.

The `s' structure contains the 6--vector $\{x y z \dot{x} ,\dot{y} \dot{z}\}$


This routine converts stars from the old, Bessel-Newcomb, FK4
system to the new, IAU 1976, FK5, Fricke system, using the method
of Standish (1982) as implemented by the RGO Nautical Almanac
office.  The NAO numerical values, as published on page x of the
1985 Astronomical Almanac, are used canonically.  There is
currently (April 1986) some controversy about whether this
algorithm is rigorously consistent with the FK4, the area of
doubt being the effect of differential e-terms on the proper
motions.

Note that conversion from Besselian epoch 1950.0 to
Julian epoch 2000.0 only is provided for.  Conversions
involving other epochs will require use of the appropriate
precession routines before and after this routine is called.

Note added in July 1998: the mean errors (and covariance matrix)
are not transformed in this program; use the general tr_ss program
to compute the transfomrations of error matrices.

References:
\begin{itemize}
\item	Standish E.M., 1982: Astron. Astrophys {\bf 115}, 20.
\item	Kaplan G.H., 1981: USNO circular {\bf 163}, A2.
\item	Yokoyama K., 1983(?): {\em The guide line in the reduction of
	observations for optical astrometry during the main campaign
	in Project Merit}, International Latitude Observatory of
	Mizusawa Report.
\item	Aoki S. et al, 1983: Astron. Astrophys. {\bf 128}, 263.
\item	Astronomical Almanac, 1985, page x.
\end{itemize}

\end{TeX}

-----------------------------------------------------------------------*/

#include <astropos.h>
#include <trigo.h>

static double A[6] = {		/* For e-term	*/
	-1.62557e-6,	-0.31919e-6,	-0.13843e-6,	
	 1.244e-3,    	-1.579e-3,   	-0.660e-3    };

static double EM[6][6] = {	/* Rotation of u + v vector	*/
  { 	 0.9999256782e0, 	-0.0111820611e0, 	-0.0048579477e0, 
		 2.42395018e-6,		-0.02710663e-6,		-0.01177656e-6},
  {	 0.0111820610e0,     	 0.9999374784e0,     	-0.0000271765e0,
		 0.02710663e-6,      	 2.42397878e-6,      	-0.00006587e-6},
  {	 0.0048579479e0,     	-0.0000271474e0,     	 0.9999881997e0,
		 0.01177656e-6,      	-0.00006582e-6,      	 2.42410173e-6},
  {	-0.000551e0,         	-0.238565e0,         	 0.435739e0,
		 0.99994704e0,       	-0.01118251e0,       	-0.00485767e0},
  {	 0.238514e0,         	-0.002667e0,         	-0.008541e0,
		 0.01118251e0,       	 0.99995883e0,       	-0.00002718e0},
  {	-0.435623e0,         	 0.012254e0,         	 0.002117e0,
		 0.00485767e0,       	-0.00002714e0,       	 1.00000956e0}
  	};

static double EM1[6][6] = {	/* FK5 ==> FK4	*/
  {   0.9999256795e0,      0.0111814829e0,     0.0048590038e0,
	      -2.42389840e-6,      -0.02710545e-6,     -0.01177742e-6},
  {  -0.0111814828e0,      0.9999374849e0,    -0.0000271771e0,
	       0.02710545e-6,      -2.42392702e-6,      0.00006585e-6},
  {  -0.0048590040e0,     -0.0000271558e0,     0.9999881946e0,
	       0.01177742e-6,       0.00006585e-6,     -2.42404995e-6},
  {  -0.000550e0,          0.238509e0,        -0.435613e0,
  	       0.99990432e0,        0.01118145e0,       0.00485852e0},
  {  -0.238559e0,         -0.002668e0,         0.012254e0,
	      -0.01118145e0,        0.99991613e0,      -0.00002717e0},
  {   0.435730e0,         -0.008541e0,         0.002116e0,
	      -0.00485852e0,       -0.00002716e0,       0.99996684e0}
  	};
static double EM2[6][6];

/*============================================================================*/
int fk4_5q (qb, vb, qj, vj)
/*+++
.PURPOSE Compute J2000 (FK5) position + velocity vector from B1950 (FK4) 
	position + velocity vector.
.METHOD  See comments above.
.RETURNS 1
---*/
  double qb[2];	/* IN: Position at epoch + equinox B1950	*/
  double vb[3]; 	/* IN: Space motion in equinox	B1950	*/
  double qj[2];	/* OUT: Position at epoch + equinox J2000	*/
  double vj[3]; 	/* OUT: Space motion in equinox J2000	*/
{ 
  struct spos s;
  double v0[3];
    /* V2.2: accept a NULL velocity */
    if (!vb) { vb = v0; fk4_qv0(qb, vb); }
    tr_qvs(qb, vb, &s);		/* Transform to structure	*/
    fk4_5s(&s, &s);		/* Perform change on structures	*/
    return(tr_sqv(&s, qj, vj));
}

/*============================================================================*/
int fk4_5s (sb, sj)
/*+++	
.PURPOSE Compute J2000 (FK5) unit-position + velocity vector 
	from B1950 (FK4) unit-position + velocity vector.
.METHOD  See comments above.
.RETURNS _COO_FK5_
---*/
  struct spos *sb; 	/* IN: Position in Phase Space B1950	*/
  struct spos *sj; 	/* OUT: Position in Phase Space J2000	*/
{ 
  double v1[6], v2[6];	/* 6-vectors	*/
  double  w, dw;
  int 	  i, j;

	/* Get 6-vector: In Fricke System, pm in "/century	*/

    for (i=3; i<6; i++)	v1[i] = sb->u[i] * (ARCSEC*1.e2);

	/* Compute scalars u.A	*/

    w 	= sb->u[0]*A[0] + sb->u[1]*A[1] + sb->u[2]*A[2];
    dw 	= sb->u[0]*A[3] + sb->u[1]*A[4] + sb->u[2]*A[5];

	/* Compute 6-vectors	*/

    for (i=0; i<3; i++)	v1[i] = sb->u[i] +  w*sb->u[i]   - A[i];
    for (i=3; i<6; i++)	v1[i] =    v1[i] + dw*sb->u[i-3] - A[i];

#ifdef DEBUG
	printf("v1 : (%20.15f,%20.15f,%20.15f)\n",
		v1[0], v1[1], v1[2]);
	printf("   : (%20.15f,%20.15f,%20.15f)\n",
		v1[3], v1[4], v1[5]);
#endif

	/* Rotate 6-vector	*/

    for (i=0; i<6; i++) {
  	w = 0.e0;
	for (j=0; j<6; j++)	w += EM[i][j] * v1[j];
	v2[i] = w;
    }

#ifdef DEBUG
	printf("v2 : (%20.15f,%20.15f,%20.15f)\n",
		v2[0], v2[1], v2[2]);
	printf("   : (%20.15f,%20.15f,%20.15f)\n",
		v2[3], v2[4], v2[5]);
#endif

	/* Normalize 6-vector	*/

    w = hypot(hypot(v2[0], v2[1]), v2[2]);
    for (i=0; i<6; i++)	 v2[i] /= w;

	/* Output new position + Derivative	*/

    for (i=0; i<3; i++)	sj->u[i] = v2[i];
    for (i=3; i<6; i++)	sj->u[i] = v2[i] / (1.e2*ARCSEC);

	/* Set up other elements of spos struct	*/
	
    sj->equinox = sj->epoch = 2000.; 
    sj->frame = _COO_FK5_ ;
    sj->precision = sb->precision ;
    sj->flags = 0 ;

    return(_COO_FK5_);
}

/*============================================================================*/
int fk4_qv0 (qb, vb)
/*+++
.PURPOSE Compute velocity vector, assuming zero velocity in FK5 
	system.
.METHOD  Approximative
.RETURNS 1
.REMARKS Useful if proper motion not known in FK4 system.
---*/
  double qb[2];	/* IN: Position at equinox B1950	*/
  double vb[3]; /* OUT: Space motion in equinox	B1950	*/
{ 
  double q[2];
  struct spos s; 
    tr_ou(qb, s.u);		/* Transform to structure	*/
    fk4_s0(&s);			/* Compute derivative		*/
    return(tr_sqv(&s, q, vb));
}

/*============================================================================*/
int fk4_s0 (sb)
/*+++	
.PURPOSE Compute velocity vector, assuming zero velocity in FK5 
	system.
.METHOD  Approximative
.RETURNS 1
.REMARKS Useful if proper motion not known in FK4 system.
---*/
  struct spos *sb; 	/* MOD: Position in Phase Space B1950	*/
{ 
  double w, dw, v1[6];
  int 	 i, j, k;

#if 0		/* ============= Rigorous Method ============== */
	/* Compute scalars u.A	*/

    w 	= sb->u[0]*A[0] + sb->u[1]*A[1] + sb->u[2]*A[2];
    dw 	= sb->u[0]*A[3] + sb->u[1]*A[4] + sb->u[2]*A[5];

	/* Compute 6-vectors, assuming zero-motion in FK4 */

    for (i=0; i<3; i++)		v1[i] = sb->u[i] +  w*sb->u[i]   - A[i];
    for (i=3; i<6; i++)		v1[i] =            dw*sb->u[i-3] - A[i];

	/* Rotate 6-vector	*/

    for (i=0; i<6; i++) {
  	w = 0.e0;
	for (j=0; j<6; j++)	w += EM[i][j] * v1[j];
	v2[i] = w;
    }

	/* Normalize 6-vector, put zero as velocities in FK5	*/

    w = hypot(hypot(v2[0], v2[1]), v2[2]);
    for (i=0; i<3; i++)	 v2[i] /= w;
    for (i=3; i<6; i++)	 v2[i] = 0.e0;

	/* Rotate back 6-vector; the result IS normalized	*/

    for (i=0; i<6; i++) {
  	w = 0.e0;
	for (j=0; j<3; j++)	w += EM1[i][j] * v2[j];
	v1[i] = w;
    }

	/* Compute scalars u.A	*/

    w 	= v1[0]*A[0] + v1[1]*A[1] + v1[2]*A[2];
    dw 	= v1[0]*A[3] + v1[1]*A[4] + v1[2]*A[5];

	/* Include A--term	*/
	
    for (i=3; i<6; i++)	v1[i] = v1[i] - dw*v1[i-3] + A[i];
    for (i=0; i<3; i++)	v1[i] = v1[i] -  w*v1[i]   + A[i];
	
	/* Normalize 6-vector	*/

    w = hypot(hypot(v1[0], v1[1]), v1[2]);
    for (i=0; i<6; i++)	 v1[i] /= w;

	/* Output Derivative	*/

    for (i=3; i<6; i++)	sb->u[i] = v1[i] / (ARCSEC*1.e2);

#else		/* ============= Quite reliable Method ============== */
		/* To speed up the computations, the matrix
		   EM2 = EM1 * (EM restricted to 3x3) is
		   used
	    	*/
    if (EM2[0][0] == 0.0)   	/* Compute the matrix EM1 * EM */
      for (i=0; i<6; i++) for (j=0; j<6; j++) {
     	w = 0;
     	for (k=0; k<3; k++) w += EM1[i][k] * EM[k][j];
     	EM2[i][j] = w;
    }
    for (i=0; i<6; i++) {
  	w = 0.0e0;
	for (j=0; j<3; j++)	w += EM2[i][j] * sb->u[j];
	v1[i] = w;
    }

	/* Compute scalars u.A	*/

    dw 	= v1[0]*A[3] + v1[1]*A[4] + v1[2]*A[5];

	/* Include A--term	*/
	
    for (i=3; i<6; i++)	v1[i] = v1[i] - dw*v1[i-3] + A[i];
	
	/* Output Derivative	*/

    for (i=3; i<6; i++)	sb->u[i] = v1[i] / (ARCSEC*1.e2);
#endif

    return(1);
}

/*============================================================================*/
int fk5_4q (qj, vj, qb, vb)
/*+++	
.PURPOSE Compute B1950 (FK4) unit-position + velocity vector
	from J2000 (FK5) unit-position + velocity vector 
.METHOD  See comments above.
.RETURNS 1
---*/
  double qj[2];	/* IN: Position at epoch + equinox J2000	*/
  double vj[3]; 	/* IN: Space motion in equinox J2000	*/
  double qb[2];	/* OUT: Position at epoch + equinox B1950	*/
  double vb[3]; 	/* OUT: Space motion in equinox	B1950	*/
{ 
  struct spos s;
    tr_qvs(qj, vj, &s);		/* Transform to structure	*/
    fk4_5s(&s, &s);		/* Perform change on structures	*/
    return(tr_sqv(&s, qb, vb));
}

/*============================================================================*/
int fk5_4s (sj, sb)
/*+++	
.PURPOSE Compute B1950 (FK4) unit-position + velocity vector
	from J2000 (FK5) unit-position + velocity vector 
.METHOD  See comments above.
.RETURNS 1
---*/
  struct spos *sj; 	/* IN: Position in Phase Space J2000	*/
  struct spos *sb; 	/* OUT: Position in Phase Space B1950	*/
{ 
  double v1[6], v2[6];	/* 6-vectors	*/
  double  w, dw;
  int 	  i, j;

	/* Get 6-vector: In Fricke System, pm in "/century	*/

    for (i=0; i<3; i++) 	v1[i] = sj->u[i];
    for (i=3; i<6; i++)	v1[i] = sb->u[i] * (1.e2*ARCSEC);

#ifdef DEBUG
	printf("v1 : (%20.15f,%20.15f,%20.15f)\n",
		v1[0], v1[1], v1[2]);
	printf("   : (%20.15f,%20.15f,%20.15f)\n",
		v1[3], v1[4], v1[5]);
#endif
	/* Rotate 6-vector	*/

    for (i=0; i<6; i++) {
  	w = 0.e0;
	for (j=0; j<6; j++)	w += EM1[i][j] * v1[j];
	v2[i] = w;
    }

	/* Normalize 6-vector	*/

    w = hypot(hypot(v2[0], v2[1]), v2[2]);
    for (i=0; i<6; i++)	 v2[i] /= w;

	/* Compute scalars u.A	*/

    w 	= v2[0]*A[0] + v2[1]*A[1] + v2[2]*A[2];
    dw 	= v2[0]*A[3] + v2[1]*A[4] + v2[2]*A[5];

	/* Include A--term	*/
	
    for (i=3; i<6; i++)	v2[i] = v2[i] - dw*v2[i-3] + A[i];
    for (i=0; i<3; i++)	v2[i] = v2[i] -  w*v2[i]   + A[i];
	
	/* Normalize 6-vector	*/

    w = hypot(hypot(v2[0], v2[1]), v2[2]);
    for (i=0; i<6; i++)	 v2[i] /= w;

	/* Output new position + Derivative	*/

    for (i=0; i<3; i++)	sb->u[i] = v2[i];
    for (i=3; i<6; i++)	sb->u[i] = v2[i] / (1.e2*ARCSEC);

#ifdef DEBUG
	printf("v2 : (%20.15f,%20.15f,%20.15f)\n",
		v2[0], v2[1], v2[2]);
	printf("   : (%20.15f,%20.15f,%20.15f)\n",
		v2[3], v2[4], v2[5]);
#endif

	/* Set up other elements of spos struct	*/
	
    sj->equinox = sj->epoch = BYtoJY(1950.) ;
    sj->frame = _COO_FK4_ ;
    sj->precision = sb->precision ;
    sj->flags = 0 ;

  return(_COO_FK4_);
}

