/*++++++++++++++
.IDENTIFICATION otype.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    
.KEYWORDS       SIMBAD Object Classification
.VERSION  1.0   04-Sep-1992
.VERSION  1.1   29-Mar-1993: External file to get Types
.VERSION  1.2   13-Apr-1993: Added ot1match, otinclass
.VERSION  1.3   27-Apr-1993: Option in ot2a
.VERSION  1.4   28-Jun-1993: Decode bad type => return #Numeric
.VERSION  1.5   21-Jul-1993: Option "E" for Extended (compatibility with old)
.VERSION  1.6   08-Dec-1993: Allow verbose types (within quotes)
.VERSION  1.7   08-Nov-1994: OTFILE macro defines the file which contains
		the definition
.VERSION  1.8   21-May-1997: Load definitions only once...
.VERSION  1.9   15-Nov-2006: OTFILE is $HOME/lib/otype.def then 
                                  /usr/local/lib/otype.def
.VERSION  2.0   09-Sep-2010: Read from http
.VERSION  2.1   30-Sep-2010: Compare just non-spaces
.VERSION  2.11  15-Oct-2010: Pb with duplicacted object types
.VERSION  2.12  07-Feb-2012: bug in load_ofile
.VERSION  2.13  13-Aug-2013: Pb in otmerge if different levels
.VERSION  2.14  27-Sep-2013: otmerge: ignore candidates
.VERSION  2.15  01-Oct-2013: Problem with Candidate branch
.COMMENTS       Routines dealing with Object Classification
		Allowed compatibility are described by two types (a0, a1).
		This must be understood as if R and r   are 2 classes
		   ir R subclass of a0 AND r overclass of a1 ==> Match
 Routines Summary:
      int otinit(char *filename)
		initialization of Object Types
    OTYPE otmerge(OTYPE t1, OTYPE t2) 
		combine two Object Types
    OTYPE a2ot(char *string)
		interprets a string as an Object Type
    OTYPE ottrunc(OTYPE t, int level)
    		truncates object type to level 0-5
    char *ot2a(OTYPE ot, char *opt)
		edit an Object Type; opt "n" (numeric) "s" (short)
      int otinclass(OTYPE t0, OTYPE t1)  
		returns 1/0 when t1 is subclass of t0
      int ot1match(OTYPE t0, OTYPE t1)  
		returns 1/0 when t1 is class or subclass of t0
      int otmatch(OTYPE t1, OTYPE t2)  
		returns 1/0 for matching/mismatching
      int otcmp(OTYPE *t1, OTYPE *t2)  
		compares two Object Types for Sorting
---------------*/
#undef DEBUG  /* utilise pour pgm main !! */
 
#include <otype.h>
#include <ctype.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifndef _ARGS
#ifdef __STDC__
#define _ARGS(A)      A   /* For ANSI, define as    A */
#else
#define _ARGS(A)      ()  /* For ANSI, define as    A */
#endif
#endif
#define starts_by(s,t)  (strncmp(s,t,sizeof(t)-1) == 0)
typedef int (*SORT_FCT)_ARGS((const void *, const void *)) ;

static char *otypes_def;	/* Default File with Definitions */

/*typedef unsigned int OTYPE;*/
typedef struct { OTYPE value;
		 char *str[3];	/* Normal /Short / Verbose definitions */
     } OTDEF;
typedef struct { OTYPE value;
		 char *text;
     } OCLASS;

/* The sum of the bits should equal 32 ... */
#define B0      4       /* Wavelength   */
/*--------------------------------------*/
#define B1      4       /* Main Class   */
#define B2      5       /* Subclass     */
#define B3      4       /* Nova/SN      */
#define Bf      15      /* NotYetUsed   */
#define Benum   (B1+B2+B3+Bf) 
#define Ball    (B0+Benum) 
#if Ball != 32
#define BAD_OTYPE Ball
#endif
/*#define is_candidate(t) ((t)&Nlev[1]) == (10<<Benum)*/
#define in_candidate_branch(t) ((t)&Nlev[1]) == (10<<Benum)

static char bits[] = { B0, B1, B2, B3, Bf };	/* Bits of each level */
static OTYPE Nt[] = { (1<<B0)-1, (1<<B1)-1, (1<<B2)-1, (1<<B3)-1, 
	(1<<Bf)-1 };
static OTYPE Nlev[] = {0, 0xf0000000, 0xff000000, 	/* Levels 0 1 2 */
		0xfff80000, 0xffff8000, 0xffffffff};	/* Levels 3 4 5 */
static int otc[sizeof(bits)];

static struct Sotypes {
	int  bytes;	/* Loaded file otype.def  */
	int  ocount;	/* Number of Object Types */
	int  scount;	/* How many synonyms	  */
	int  acount;	/* How many Allowed equiv */
	char *text;	/* Loaded file		  */
	OTDEF  *classes;	/* The classes	  */
	OCLASS *symbols;	/* Class synonyms */
	OTYPE  *allowed;	/* Allowed	  */
  } otypes;
static char check_numeric;	/* Used by ntype to convert numeric to otype */
  
#define ITEMS(s)	(sizeof(s)/sizeof(s[0]))
#ifndef O_BINARY
#define O_BINARY 0
#endif
	/* Functions */
OTYPE a2ot(), otmerge(); char *ot2a();

/*===========================================================================*/

static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
  for ( s = text; *s && (*s != c); s++) ;
  return(s-text);
}

static int words(s)
/*++++++
.PURPOSE Computes Number of Words (separated by blanks)
.RETURNS Number of Words (note the quote defines a single word)
--------*/
  char *s;	/* IN: String to compare */
{
  char *p; int nw;
    for (p = s, nw=0; *p; ) {
    	while(isspace(*p)) p++;
	if (*p == '"') { 
	    nw++; p++;
	    p += strloc(p, '"');
	    if (*p) p++;
	    continue;
	}
    	if (isgraph(*p)) { nw++; while(isgraph(*p)) p++; }
    }
    return(nw);
}

static char *set_oclass(poc, str)
/*++++++
.PURPOSE Set the OCLASS pointer with the proper text, with quotes removed.
.RETURNS Pointer to next word.
--------*/
  OCLASS *poc;		/* MOD: OCLASS structure to set up */
  char   *str;		/* IN: String with ascii equivalent */
{
    while (isspace(*str)) str++;	/* Skip leading blanks */
    poc->text = str;
    if (*str == '"') {
	poc->text = ++str;
	str += strloc(str, '"');
    	if (*str) *(str++) = 0;
    }
    else  while (isgraph(*str)) str++;
    while (isspace(*str)) *(str++) = 0;
    return(str);
}

/*#define is_candidate(t) ((t)&Nlev[1]) == (10<<Benum)*/
static int studiff(s1, s2)
/*++++++
.PURPOSE Computes the difference of two strings, case independent
.RETURNS The Difference s1 - s2
--------*/
  char *s1, *s2;	/* IN: Strings to compare */
{
  int ds;
    for (ds = 0; ds == 0; s1++, s2++) {
	ds = toupper(*s1) - toupper(*s2);
	if (!*s1) break;
    }
    return(ds);
}

#if 0
static int studiff1(s1, s2)
/*++++++
.PURPOSE Computes the difference of 1st word of 2 strings, case-independent
.RETURNS The Difference s1 - s2
--------*/
  char *s1, *s2;	/* IN: Strings to compare */
{
  int ds;
    for (ds=0; (ds==0) && isgraph(*s1) && isgraph(*s2); s1++, s2++) {
	ds = toupper(*s1) - toupper(*s2);
    }
    return(ds);
}
#endif

static char *ot2an(ot)
/*++++++++++++++++
.PURPOSE  Convert object type into its string, expressed in Numeric
.RETURNS  Pointer to character equivalence / NULL
.REMARKS  See also ot2a
-----------------*/
  OTYPE ot;	/* IN: Object type to interpret */
{
  static char buf[20];	/* Only 12 used... */
  OTYPE otyp; int i;

  /* if (!otypes.bytes) load_otypes(otypes_def);	-- Load Definitions */
    for (i=ITEMS(bits) - 1, otyp = ot >> Bf; --i >= 0; otyp >>= bits[i]) 
	otc[i] = otyp & Nt[i];
    sprintf(buf, "%02d.%02d.%02d.%02d", otc[0], otc[1], otc[2], otc[3]);
    return(buf);
}

static int ot1diff(cl1, cl2)
/*++++++++++++++++
.PURPOSE  Compare two object types on the linear scale
.RETURNS  difference
.REMARKS  for qsort / bsearch
-----------------*/
  OCLASS *cl1, *cl2;
{
  if(cl1->value < cl2->value) return(-1);
  return(cl1->value != cl2->value);
}

static int osdiff(cl1, cl2)
/*++++++++++++++++
.PURPOSE  Compare two object types symbols
.RETURNS  difference
.REMARKS  for qsort / bsearch
-----------------*/
  OCLASS *cl1, *cl2;
{
  int dif;
    dif = studiff(cl1->text, cl2->text);
    return(dif);
}

static int is_candidate(OTYPE ot)
/*++++++
.PURPOSE Returns non-zero if type is a candidate
.RETURNS 0=non-candidate 1=in Candidate branch 2=in generic branch -1=bad
--------*/
{
  OTDEF  cl, *class;
    if((ot&Nlev[1]) == (10<<Benum)) return(1);
    cl.value = ot; 
    class = (OTDEF *)bsearch(&cl, otypes.classes, otypes.ocount,
		                sizeof(OTDEF), (SORT_FCT)ot1diff);
    if (!class) return(-1);
    if (strchr(class->str[1], '?')) return(2);
    return(0);
}

#if 0
static int osdiff1(cl1, cl2)
/*++++++++++++++++
.PURPOSE  Compare two object types symbols (only first symbol)
.RETURNS  difference
.REMARKS  for qsort / bsearch
-----------------*/
  OCLASS *cl1, *cl2;
{
    return(studiff1(cl1->text, cl2->text)) ;
}
#endif

/*===========================================================================*/
static OTYPE ntype(s)
/*++++++
.PURPOSE Convert the set of Numbers into Class
.RETURNS The interpreted number / -1 if error
--------*/
  char *s;     /* IN: Class to convert */
{
	/* For numeric type, assume the hierarchical numeric scheme */
  OTYPE ot; int i, n; 
    ot = 0;
    for (i=0; (*s) && (i < sizeof(bits)); i++) {
	ot <<= bits[i]; 
	if (*s == '.') s++;
	n = atoi(s); 
	if (n >= (1L << bits[i])) return((OTYPE)-1);
	ot |= n; 
	while (isdigit(*s)) s++;
    }
    while (i < sizeof(bits)) ot <<= bits[i++]; 

    	/* Check the type is existing */
    if (check_numeric) { OTDEF  cl;
    	cl.value = ot; 
    	if (! bsearch(&cl, otypes.classes, otypes.ocount, 
    	    sizeof(OTDEF), (SORT_FCT)ot1diff)) ot = (OTYPE)(-1);
    }
    return(ot);
}

static int load_ofile(struct Sotypes *otypes, char *name)
/*++++++
.PURPOSE Load the data from a URL or a file
.RETURNS The number of bytes (otypes->bytes)
--------*/
{
  static char wget[] = "wget -q -O - '%s'";
  int bytes;
  FILE *fp;

    memset(otypes, 0, sizeof(struct Sotypes));
    if (starts_by(name, "http://")) { char *cmd, *abuf=0, *ebuf, *cbuf=0; 
        cmd = malloc(strlen(name) + sizeof(wget)); sprintf(cmd, wget, name);
        if ((fp = popen(cmd, "r"))) { 
	    abuf = cbuf = malloc(20<<10); ebuf = abuf + (20<<10);
	    while ((bytes = fread(cbuf, 1, (ebuf-cbuf), fp))>0) {
	        cbuf += bytes;
	        if ((ebuf-cbuf)<16) {
		    int n = (ebuf-abuf) + (4<<10);	/* Add 4Kb */
		    int o = (cbuf-abuf);
		    abuf = realloc(abuf, n);
		    ebuf = abuf + n;
		    cbuf = abuf + o;
	        }
	    }
	    *cbuf = 0;		/* Terminate buffer */
	    pclose(fp);
	    if ((cbuf-abuf)<1000) { /* Too short ! */
	        fprintf(stderr, "#***otype: %s => %d bytes\n", 
			name, (int)(cbuf-abuf));
	        free(abuf); abuf = ebuf = (char *)0;
	    }
        }
        free(cmd);
	if (abuf) {
            otypes->text  = abuf;
	    otypes->bytes = (cbuf-abuf);
	}
    }
    else if ((fp = fopen(name, "r"))) {	/* A standard file */
	fseek(fp, 0, SEEK_END); bytes = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	otypes->text  = malloc(bytes+1);
	bytes = fread(otypes->text, 1, bytes, fp);
	otypes->text[bytes] = 0;
	otypes->bytes = bytes;
    }
    return(otypes->bytes);
}

static void load_otypes(char *file /* File otypes.def */)
/*++++++
.PURPOSE Load the Object Types into Memory
.RETURNS (Exit(1) if Error)
--------*/
{
static char url[]="\
http://simbad.u-strasbg.fr/simbad/sim-display?data=otypes&form=plain";
static char err_msg[] = "\
The file 'otype.def' (list of SIMBAD object types) could not be found,\n\
neither from URL, nor in $HOME/lib or /usr/local/lib. The tested URL is:\n\
";
  int i, jc, js, k, lineno, status; 
  char *pc, *pe, *p, oloaded;
  OCLASS  *s, *e;
  OTYPE ot, *a;
  OTDEF   *o;
#ifdef DEBUG
    int is;
#endif

    /* Verify: already loaded ?  */
    if (otypes.bytes) {
	fprintf(stderr, 
	"#***Trying to load '%s' for otypes a second time (had %d bytes)\n",
	file, otypes.bytes) ;
	return ;
    }

    /* If file is empty, try to locate it */
    if (!file) file = "";
    if (*file) 
	load_ofile(&otypes, file);
    else {
	if ((p = getenv("OTFILE"))) {
	    load_ofile(&otypes, file=p);
	    if (otypes.bytes<=0) fprintf(stderr, 
		"#***otype: $OTFILE=%s unreadable!\n", p);
	}
	if (otypes.bytes<=0)  	/* Try from URL */
	    load_ofile(&otypes, file=url);
	if ((otypes.bytes<=0) && (p = getenv("HOME"))) { /* Try from home */
	    file = malloc(strlen(p)+16);
	    strcpy(file, p); strcat(file, "/lib/otype.def");
	    load_ofile(&otypes, file);
	    if (otypes.bytes<=0) fprintf(stderr,
		"#***otype: missing file: %s ***\n", file);
	    free(file);
	    file = (char *)0;
	}
	if (otypes.bytes<=0)  	/* Try /usr/local/lib/otype.def */
	    load_ofile(&otypes, file="/usr/local/lib/otype.def");
    }
    if (otypes.bytes<=0) {
	fprintf(stderr, "#***"); perror(file); 
	fprintf(stderr, "%s\n", err_msg);
	fprintf(stderr, "%s\n", url);
	exit(1); 
    }
    fprintf(stderr, "#...object types from: %s\n", file);

    /* file now in Memory */
    otypes.text[otypes.bytes] = 0;

#ifdef DEBUG
    printf("Loaded from file %s:\n%s\n", file, otypes.text);
#endif

    /* Count number of Classes (ocount) (1/line) and Symbols */
    otypes.ocount = 1;	/* Include BAD_OTYPE */
    for (js=0, pc=otypes.text, pe = pc+otypes.bytes; pc < pe; pc += 1+i) {
        if (*pc == '+') pc++ ;		/* Special */
        if (*pc == '*') pc++ ;		/* Special */
        i = strloc(pc, '\n');
        if (i == 0) continue;		/* Emtpy line */
        if (*pc == '#') continue;	/* Comment */
        if (*pc == '-') continue;	/* Ignored */
        if (*pc == '(') continue;	/* Ignored */
	pc[i] = 0;
	for (p = pc; isspace(*p); p++) ;
	jc = strloc(pc, ':');
	if(pc[jc]) {		/* Definition or Synonym */
	    if (isdigit(*p)) { 
	    	otypes.ocount += 1; 
	    	otypes.scount += words(pc+jc+1); 
	    }
	    else js++;
	}
	else if (pc[strloc(pc, '/')]) otypes.acount += 1; /* Allowed */
	pc[i] = '\n';
    }
	/* Allocate the table */
#ifdef DEBUG
	printf("ocount=%d, scount=%d, acount=%d\n", otypes.ocount,
		otypes.scount, otypes.acount);
	is = 0;
#endif
    otypes.classes = (OTDEF  *)malloc((otypes.ocount)*sizeof(OTDEF ));
    otypes.symbols = (OCLASS *)malloc((otypes.scount+js)*sizeof(OCLASS));
    otypes.allowed = (OTYPE  *)malloc((otypes.acount*2 )*sizeof(OTYPE));

	/* Fill the otypes */
    status = 0; ot = 0; oloaded = 0;
    for (lineno=1, o = otypes.classes, s = otypes.symbols, a = otypes.allowed,
        pc = otypes.text, pe = pc+otypes.bytes;
        pc < pe; pc += 1+i, lineno++) {
        if (*pc == '+') pc++ ;		/* Special 	*/
        if (*pc == '*') pc++ ;		/* Special 	*/
        i = strloc(pc, '\n');
        pc[i] = 0;              	/* Remove \n	*/
        if (*pc == '#') continue;	/* Comment 	*/
        if (*pc == '-') continue;	/* Ignored 	*/
        if (*pc == '(') continue;	/* Ignored 	*/
	jc = strloc(pc, '#');	/* Check strange configurations */
	if (pc[jc] && (jc < 12)) 
	    fprintf(stderr, "++++Is the Strange line #%d a comment?:\n%s\n", 
		lineno, pc);
	for (p = pc+i-1; isspace(*p); p-- ) *p = 0;	  /* Remove tail */
	for (p = pc; isspace(*p); p++) ;		  /* Skip   head */
        if (!*p)       continue;       	/* Empty lines  */
	jc = strloc(p, ':');
	if (p[jc] && isdigit(*p)) {			/* Type Number	*/
	    p[jc] = 0; 
	    o->value = ntype(p); 
	    if (o->value == (OTYPE)-1)
	    	status = -1,
		fprintf(stderr, "#*** Bad class file %s, line %d <%s>\n", 
			file, lineno, p);
	    if (ot && o->value <= ot)
	    	status = -1,
		fprintf(stderr, "#*** Unsorted class file %s, line %d <%s>\n", 
			file, lineno, pc);
	    ot = o->value;
	    for (p += jc+1; isspace(*p); p++) ;
	    o->str[0] = o->str[1] = o->str[2] = p;
	    while(*p) {
		if (*p == '"') k = 2;
		else if (o->str[0] == o->str[1]) k = 1;
		else k = 0;
	    	s->value = o->value; 
		p = set_oclass(s, p);	/* List sorted by symbol */
	    	if (k) o->str[k] = s->text;
#ifdef DEBUG
		is++;
#endif
		s++;
	    }
	    o++; continue;
	}
#ifdef DEBUG
	printf(".... Dealt with scount %d\n", is);
#endif
	if (! oloaded) {		/* Add BAD_OTYPE */
	    o->value = (OTYPE)-1, o->str[1] = "#?#";
	    o->str[0] = o->str[2] = "##BAD_type##";
	    oloaded = 1;
	    qsort(otypes.symbols, otypes.scount, sizeof(OCLASS), 
		(SORT_FCT)osdiff);
	}
	if (p[jc]) {			/* Synonym	*/
	    p[jc] = ' '; 
	    p = set_oclass(s, p);
	    s->value = a2ot(p);
	    if (s->value == (OTYPE)-1) status = -1,
		fprintf(stderr, "#*** Unknown class file %s, line %d <%s>\n", 
			file, lineno, p);
	    s++; continue;
    	}
	jc = strloc(p, '/');
	if (p[jc]) {			/* Allowed	*/
	    p[jc] = ' '; while(isspace(p[jc])) jc++;
	    ot = a2ot(p+jc);		/* 2nd number	*/
	    if (ot == (OTYPE)-1) status = -1,
		fprintf(stderr, "#*** Unknown class file %s, line %d <%s>\n", 
			file, lineno, p+jc);
	    while(isspace(p[--jc])) p[jc] = 0;
	    *a = a2ot(p);
	    if (*a == (OTYPE)-1) status = -1,
		fprintf(stderr, "#*** Unknown class file %s, line %d <%s>\n", 
			file, lineno, p);
	    a[1] = ot;
	    a += 2; continue;
	}
    }
    if (status) exit(1);
    otypes.scount += js;
    qsort(otypes.symbols, otypes.scount, sizeof(OCLASS), (SORT_FCT)osdiff);

    	/* Check non-ambiguous symbols, remove duplicates */
    for (s=otypes.symbols, e=s+(otypes.scount-1); s<e; s++) {
	if (studiff(s[1].text, s[0].text)) continue;
	/* Identical symbol... */
	if (s[1].value != s[0].value) {
	    fprintf(stderr, "#*** Ambiguous classes: %s=%s, ",
		s[1].text, ot2an(s[1].value));
	    fprintf(stderr, "%s=%s\n", s[0].text, ot2an(s[0].value));
	}
	else {
#ifdef TEST
	    fprintf(stderr, "#...Duplication: %s=%s\n", 
		    s[0].text, ot2an(s[0].value));
#endif
	    memmove(s, s+1, (e-s)*sizeof(OCLASS));
	    --e;
	    otypes.scount -= 1;
	}
    }
    check_numeric = 1;	/* Check further translation of numeric otype */
    return;
}

static int level(ot)
/*++++++++++++++++
.PURPOSE  Find out the level of the classification
.RETURNS  Level between 0(unknown) and  ...
-----------------*/
  OTYPE ot;	/* IN: Type to check */
{
  int i;
#if 0
     for (i=0; ot; ot <<= bits[i++]) ;
     return(i);
#else	/* V2.13: works on all architectures */
     for (i=0; (i<ITEMS(Nlev)) && ((ot&(~Nlev[i]))!=0) ; i++) ;
     return(i);
#endif
}

int otinclass(cl, ot)
/*++++++
.PURPOSE Check if ot is in class
.RETURNS 1 if ot equates cl, or is subclass of cl
.REMARKS Uses level
--------*/
  OTYPE cl;	/* IN: Types to compare */
  OTYPE ot;	/* IN: Types to compare */
{
  int i;
#ifdef DEBUG
    printf("...otinclass(%s", ot2a(cl, "n"));
    printf(", %s) =>", ot2a(ot, "n"));
#endif
    if ((cl|ot) == BAD_OTYPE)	/* V2.51: cl or ot is BAD_OTYPE */
	return(0);
    i = level(cl);
#ifdef DEBUG
    printf("%d\n", (ot & Nlev[i]) == cl);
#endif
    return((ot & Nlev[i]) == cl);
}

/*===========================================================================*
 *		Public Routines
 *===========================================================================*/
int otinit(file)
/*++++++
.PURPOSE Initialize (load definitions)
.RETURNS 0/-1
--------*/
  char *file;	/* IN: File to use for Object Class Definitions */
{
    if (!otypes.bytes) load_otypes(file);
    return(0);
}

OTYPE a2ot(s)
/*++++++++++++++++
.PURPOSE  Interpret a Class from a string
.RETURNS  Found type / -1 if error
.REMARKS  String may be in numeric or alphabetic format
-----------------*/
  char *s;	/* IN: String to interpret */
{
  OCLASS *class, cl;

    if (!otypes.bytes) load_otypes(otypes_def);	/* Load Definitions */
    while(isspace(*s)) s++;
    if(isdigit(*s)) return(ntype(s));
    cl.text = s;
    class = (OCLASS *)bsearch(&cl, otypes.symbols, otypes.scount,
	sizeof(OCLASS), (SORT_FCT)osdiff);
    if (class) return(class->value);
    return(-1);
}

/*===========================================================================*/

OTYPE ottrunc(t, lev)
/*++++++++++++++++
.PURPOSE  Truncate an Object Type
.RETURNS  The result...
-----------------*/
  OTYPE t;	/* IN: Type to truncate */
  int lev;	/* IN: Level, range 0-5 */
{
  int ll;

    if (t == (OTYPE)-1) return(t);
    ll = lev; 
    if (ll < 0) ll = 0;
    if (ll >= ITEMS(Nlev)) ll = ITEMS(Nlev) - 1;
    return(t & Nlev[ll]);
}

OTYPE otmerge(t1, t2)
/*++++++++++++++++
.PURPOSE  Combine two object types
.RETURNS  The result...
-----------------*/
  OTYPE t1;	/* IN: First type to merge */
  OTYPE t2;	/* IN: 2nd   type to merge */
{
  OTYPE r; int i1, i2, c1, c2;

  /* Obvious merges: */
  if ((t1==0)||(t1==BAD_OTYPE)) return(t2);
  if ((t2==0)||(t2==BAD_OTYPE)) return(t1);
  if (t1==t2) return(t1);

  if (!otypes.bytes) load_otypes(otypes_def);	/* Load Definitions */
	/* Choose the deepest level */
#ifdef DEBUG
    printf("....otmerge(%s", ot2an(t1));
    printf(", %s) => ", ot2an(t2));
#endif
    i1 = level(t1); c1 = is_candidate(t1);
    i2 = level(t2); c2 = is_candidate(t2);
#if 0 /* V2.13: condidates are degraded */
    if (is_candidate(t1)) i1--;
    if (is_candidate(t2)) i2--;
    if (i1 == i2) r = t1 > t2 ? t1 : t2;
    else r = i1 > i2 ? t1 : t2;
#else
    if (c1!=c2) {	/* Candidate with non-candidate */
	r = c1 ? t2 : t1;
    }
    else if (i1 == i2) r = t1 > t2 ? t1 : t2;
    else r = i1 > i2 ? t1 : t2;
#endif
#ifdef DEBUG
    printf("%s\n", ot2an(r));
#endif
    return(r);
}

int otcmp(cl1, cl2)
/*++++++++++++++++
.PURPOSE  Compare two object types on the "information" scale
.RETURNS  difference: -1 if cl1 is "less precise" than cl2
-----------------*/
  OTYPE *cl1; 	/* IN: First Object Type */
  OTYPE *cl2;	/* IN: 2nd  Object Type  */
{
  if(*cl1 == *cl2) return(0);
  if(otmerge(*cl1, *cl2) == *cl2) return(1);
  return(-1);
}

int ot1match(t0, t1)
/*++++++
.PURPOSE Check if t1 is class or subclass of t0
.RETURNS 1 if Match
.REMARKS 
--------*/
  OTYPE t0;	/* IN: subclass */
  OTYPE t1;	/* IN: Overclass */
{
  int i, match;
  OTYPE *a;
   
#ifdef DEBUG
    printf("----ot1match(%s", ot2an(t0));
    printf(", %s)\n", ot2an(t1));
#endif

    if ((t0|t1) == BAD_OTYPE)	/* V2.51: BAD_OTYPE never matches */
	return(0);

    	/* If t1 is subtype of t0 obvious match */
    if (otinclass(t0, t1)) return(1);

   	/* Check if however there is an agreement 
   	   This second pass provides an agreement if
   	     ocl[0] subclass of a[0] AND ocl[1] subclass of a[1]
   		for two allowed types a[0] and a[1]
	*/
    for (i = otypes.acount, a = otypes.allowed, match = 0; 
     	(!match) && (--i>=0); a += 2) {
   	if (otinclass(a[0], t1)) match = otinclass(t0, a[1]);
   }
   return(match);
}

int otmatch(t1, t2)
/*++++++
.PURPOSE Check if 2 Object Types agree
.RETURNS 1 if Match
.REMARKS 
--------*/
  OTYPE t1;	/* IN: Types to compare */
  OTYPE t2;	/* IN: Types to compare */
{
  OTYPE ocl[2]; int ok; char *p;
   
#ifdef TEST /*DEBUG*/
    printf("----otmatch(%s", ot2an(t1));
    printf(", %s)\n", ot2an(t2));
#endif

    if ((t1|t2) == BAD_OTYPE)	/* V2.51: BAD_OTYPE never matches */
	return(0);

    if (t1 == t2) return(1);	/* Obvious match ! */
    
    	/* ocl[0] is the class with the higher "information", 
    	   ocl[1] the lower.
    	   Corresponding levels are lev[0] >= lev[1]
       */
    ocl[0] = otmerge(t1, t2);
    ocl[1] = ocl[0] == t1 ? t2 : t1;
#ifdef DEBUG
    printf("........... %s", ot2an(ocl[0]));
    printf("/ %s\n", ot2an(ocl[1]));
#endif

#if 0
    	/* If ocl[0] is subtype of ocl[1], obvious match */
    if (otinclass(ocl[1], ocl[0])) return(1);
    
   	/* Check if however there is an agreement 
   	   This second pass provides an agreement if
   	     ocl[0] subclass of a[0] AND ocl[1] subclass of a[1]
   		for two allowed types a[0] and a[1]
	*/
    for (i = otypes.acount, a = otypes.allowed, match = 0; 
     	(!match) && (--i>=0); a += 2) {
   	if (otinclass(a[0], ocl[0])) match = otinclass(ocl[1], a[1]);
    }
    return(match);
#else
    if ((ok = ot1match(ocl[1], ocl[0])))
	return(ok);
    /* Verify Candidate Branch */
    if (in_candidate_branch(ocl[1])) ok = 1;
    else if (in_candidate_branch(ocl[0])) ok = 0;
    else return(ok);
#if 0
    else { char atype[20];		/* Not in "Candidates" branch;
					   try if type with a question mark?*/
	if ((p = ot2a(ocl[1], 0)))  {
	    strncpy(atype, p, sizeof(atype));
	    atype[sizeof(atype)-1] = 0;
	    p = strchr(atype, '?');
	}
	if (p) ok = 1;
	else if ((p = ot2a(ocl[0], 0)))  {
	    strncpy(atype, p, sizeof(atype));
	    atype[sizeof(atype)-1] = 0;
	    if ((p = strchr(p, '?'))) ok = 0;
	    else return(ok);
	}
	*p = 0;			/* Remove the ? */
	return(otmatch(a2ot(atype), ocl[ok^1]));
    }
#endif
    p = ot2a(ocl[ok], 0);	/* Candidate as Possible_... or Possible_... */
    if ((p = strchr(p, '_'))) 
	return(otmatch(a2ot(p+1), ocl[ok^1])) ;
    return(0);
#endif
}

char *ot2a(ot, opt)
/*++++++++++++++++
.PURPOSE  Convert object type into its string
.RETURNS  Pointer to character equivalence 
.REMARKS  Option string includes "n" for Numeric edition, "s" for Short
-----------------*/
  OTYPE ot;	/* IN: Object type to interpret */
  char *opt;	/* IN: Option string, "n" for Numeric, "s" for Short */
{
  static char *bad_type;
  OTDEF  cl, *class;
  int choice; char c, *p;

    if (!otypes.bytes) load_otypes(otypes_def);	/* Load Definitions */
    if(!opt) opt = "";

    cl.value = ot; choice = 0; p = opt;
    while(*p) switch(c = *(p++)) {
	case 'v': case 'V':	/* Verbose */
	case 'e': case 'E':	/* Extended */
	    choice = 2; continue;
	case 'd': case 'D':	/* Default */
	    choice = 0; continue;
	case 'n': case 'N':	/* Numeric */
	    choice = -1; continue;
	case 's': case 'S':	/* Short designation */
	    choice = 1; continue;
	default: 
	    if (isdigit(c)) cl.value = ottrunc(cl.value, c-'0');
	    else if (isspace(c)) continue;
	    else fprintf(stderr, 
		"##** ot2a: Bad character <%c> in option string \"%s\"\n", 
		c, opt);
	    continue;
    }

    if(choice < 0) return(ot2an(ot));
    class = (OTDEF *)bsearch(&cl, otypes.classes, otypes.ocount,
		sizeof(OTDEF), (SORT_FCT)ot1diff);
    if (class) return(class->str[choice]);
	/* Bad Otype: Edit it as NUMERIC */
    p = ot2an(ot);
    if (!bad_type) { bad_type = malloc(20+2); *bad_type = '#'; }
    strcpy(bad_type+1, p); strcat(bad_type+1, "#");
    return(bad_type);
}


#ifdef TEST
int otf(buf, f, ot)
/*++++++++++++++++
.PURPOSE  Formatted edition of the object type
.RETURNS  Number of bytes written
-----------------*/
  char *buf;    /* OUT: Output buffer */
  char *f;    /* IN: Format for edition Numeric Level Alphabetic Indent */
  OTYPE ot;	/* IN: Object type to interpret */
{
  char *p;
  int i;

    *buf = 0;
    for (p = buf; *f; f++) {
	if (*f == '%') switch(*++f) {
	  case 'L': case 'l': 	/* Level */
		*(p++) = '0' + level(ot); 
		break;
	  case 'I': case 'i':	/* Indentation */
		for (i = level(ot); --i >= 0; p += 4)
		    p[0] = p[1] = p[2] = p[3] = ' ';
		break;
	  case 'A': case 'a':	/* Alphabetic  */
	  case 'S': case 's':
		strcpy(p, ot2a(ot, "s"));
		p += strlen(p);
		break;
	  case 'E': case 'e':
	  case 'V': case 'v':
		strcpy(p, ot2a(ot, "v"));
		p += strlen(p);
		break;
	  case 'D': case 'd':	/* numeric */
	  case 'N': case 'n':	/* numeric */
		strcpy(p, ot2an(ot));
		p += strlen(p);
		break;
	   default:	/* Just copy */
		*(p++) = *f;
		break;
	}
	else *(p++) = *f;
	*p = 0;
    }
    if (p == buf) strcpy(p, ot2a(ot, ""));
    return(strlen(buf));
}

/*===========================================================================*/
void otprt()
/*++++++++++++++++
.PURPOSE  Print all object types
.RETURNS  -
-----------------*/
{  
	OTDEF *class; int i;
	char buf[80];

   if (!otypes.bytes) load_otypes(otypes_def);	/* Load Definitions */
   for (i=otypes.ocount, class = otypes.classes; --i >= 0; class++)
	otf(buf, "[%l] %n: %i%a", class->value), puts(buf);
}

int main()
/*++++++++++++++++
.PURPOSE  TEST Program
.RETURNS  ---
-----------------*/
{
	char buf[BUFSIZ], *p, x;
	OTYPE t1, t2, tt, *pt;
	OCLASS *o;
	int i, lev;

  /* otinit("otype.def"); */
  printf("Bad Type: %s = %s\n", ot2an(-1L), ot2a(-1L, ""));
  for(o = otypes.symbols, i = otypes.scount; --i >= 0; o++)
	printf("%20s = %s\n", o->text, ot2an(o->value));
  p = "V.blbbla";
  t1 = a2ot(p);
  printf("\n%19s => %s = %s\n", p, ot2an(t1), ot2a(t1, ""));

  printf("Give level for Truncate: "); 
  fgets(buf, sizeof(buf), stdin); 
  lev = iscntrl(buf[0]) ? 4 : atoi(buf);
  while(1) {
    printf("Give two Classes: ");
    if (!fgets(buf, sizeof(buf), stdin)) break;
    for (p=buf+strlen(buf)-1; isspace(*p); p--) *p = 0;
    if (iscntrl(*buf)) {
    	puts("List of All Types:"); otprt();
	puts("List of Compatible Types");
  	for(pt = otypes.allowed, i = otypes.acount; --i >= 0; pt += 2)
	    printf("%20s = ", ot2a(pt[0], "")), 
	    printf("%s\n", ot2a(pt[1], ""));
	continue;
    }
    for (p = buf; isspace(*p); p++) ; while(isgraph(*p)) p++;
    x = *p; *p = 0;
    t1 = a2ot (buf); if (x) *(p++) = x;
    t2 = a2ot (p); 
    tt = ottrunc(t1, lev);
    otf(buf, "(1) [%l] %n: %i%s %v", t1); printf("%s trunc=%s\n", buf,
    	ot2a(tt,  ""));
    tt = ottrunc(t2, lev);
    otf(buf, "(2) [%l] %n: %i%a", t2); printf("%s trunc=%s\n", buf,
    	ot2a(tt,  ""));
    otf(buf, " => [%l] %n: %i%a", otmerge(t1, t2)); 
    p = otmatch(t1, t2) ? "agree" : "DISAGREE";
    printf("%s (types %s)\n", buf, p);
  }
  exit(0);
}
#endif
