/*++++++++++++++++++++++++++++++++++++++++++
.TYPE		Module
.IDENTIFICATION	surf.c
.AUTHOR		Francois Ochsenbein [ESO]  
.VERSION  1.0 	15-Nov-1985: Creation
.VERSION  1.1	26-May-1989: Cosmetics
.VERSION  1.2	26-Feb-1992: Tested on Turbo-C
.VERSION  1.3	26-Jul-2001: More accurate formula for surface of rectangle
.LANGUAGE	C
.ENVIRONMENT	Any
.COMMENTS
\begin{TeX}
	Routines provided in this module return the surface of simple
	geometries on the sphere.
	Angles are assumed to be expressed in {\bf degrees},
        and surfaces in {\bf square degrees}.
\end{TeX}
-------------------------------------------*/ 

#include <trigo.h>
 
/* Scalar product */
#define dotprod(v1, v2)	(v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2])

/*===========================================================================*/
double surf_u3(u0, u1, u2)
/*+++++++++++++++++++++++++++++++++++++++++
.PURPOSE Compute the surface (square degrees) of a spherical triangle
.RETURNS the surface (square degrees)
.REMARKS Use l'Huilier formula:
    tan^2(S/4) = tan((a+b+c)/4) tan((a+b-c)/4) tan((a-b+c)/4) tan((b+c-a)/4)
      	 where a b c = length of triangle edges.
-------------------------------------------*/
  double u0[3] ;	/* IN: Vector defining 1st point */
  double u1[3] ;	/* IN: Vector defining 2nd point */
  double u2[3] ;	/* IN: Vector defining 3rd point */
{
  double a,b,c,t ;
    a = acos(dotprod(u1, u2)) ;
    b = acos(dotprod(u2, u0)) ;
    c = acos(dotprod(u0, u1)) ;
    t = tan((a+b+c)/4.) * tan((a+b-c)/4.) * tan((a-b+c)/4.) * tan((b+c-a)/4.);
    return(4. * DEG * DEG * atan(sqrt(t))) ;
}

double surf_p3(a,b,c)     
/*++++++++++++++++++++++++++++++++++++++++++
.PURPOSE Computation of the surface (in square degrees)
	 of a triangle on a projection plane (gnomonic projection)
         defined by   
	 - the center
         - the point (a,b)
         - the point (a,c)
.RETURNS the surface (square degrees)
-------------------------------------------*/
  double a; 	/* First position */
  double b; 	/* (a,b) provide the first position */
  double c; 	/* (a,c) is the second point */
{
  double phi1,phi2,sec, returned_value;
    if(a == 0.e0) return(0.e0);
    phi1 = atan (b/a);  phi2 = atan(c/a); sec = sqrt(1.e0+a*a);
    returned_value = phi2 - asin(sin(phi2)/sec)
                   - phi1 + asin(sin(phi1)/sec);
    return ( DEG * DEG * returned_value);
}

/*===========================================================================*/
#if 0
double surf_p(p1,p2) 
/*++++++++++++++++++++++++++++++++++++++++++
.PURPOSE Computation of the surface (in square degrees)
	of a rectangle on the projection
.RETURNS the surface (square degrees)
.REMARKS Use triangles.
-------------------------------------------*/
  double p1[2]; 	/* IN: position (X,Y) of the lower left corner */
  double p2[2]; 	/* IN: position (X,Y) of the upper right corner*/
{
  double surf_p3();
    return ( fabs(surf_p3(p2[0],p1[1],p2[1]) + surf_p3(p1[0],p2[1],p1[1]) 
                + surf_p3(p2[1],p1[0],p2[0]) + surf_p3(p1[1],p2[0],p1[0])));
}
#else
double surf_p(p0,p1) 
/*++++++++++++++++++++++++++++++++++++++++++
.PURPOSE Computation of the surface (in square degrees)
	of a rectangle on the projection
.RETURNS the surface (square degrees)
.REMARKS Use formula
   S = Arcsin(m_0.n_0) + Arcsin(m_1.n_1) - Arcsin(m_1.n_0) - Arcsin(m_0.n_1)
   where  m_i = X_i/sqrt(i+X_i^2)
   (more accurate than triangles)
   Results from   S = A_0 + A_1 + A_2 + A_3 - 2*pi
   where A_i = angles of the rectangle
   and the angle A_1 is given from
              cos(0,2) - cos(1,0) cos(1,2)
   cos(A_1) = ----------------------------
		   sin(1,0) sin(1,2)
   (+ similar formulae with circular permutations)
-------------------------------------------*/
  double p0[2]; 	/* IN: position (X,Y) of the lower left corner */
  double p1[2]; 	/* IN: position (X,Y) of the upper right corner*/
{
  double m[4];
    m[0] = p0[0]/sqrt(1.+p0[0]*p0[0]) ;
    m[1] = p0[1]/sqrt(1.+p0[1]*p0[1]) ;
    m[2] = p1[0]/sqrt(1.+p1[0]*p1[0]) ;
    m[3] = p1[1]/sqrt(1.+p1[1]*p1[1]) ;
    return(fabs(DEG * DEG * (asin(m[0]*m[1]) + asin(m[2]*m[3]) 
           - asin(m[0]*m[3]) - asin(m[1]*m[2])))) ;
}
#endif

/*===========================================================================*/
double surf_o(q1,q2)
/*++++++++++++++++++++++++++++++++++++++++++
.PURPOSE Computation of the surface (in square degrees)
	of a "rectangle" on the sphere
.RETURNS The surface (square degrees)
-------------------------------------------*/
  double q1[2];	/* IN: position (ra,dec) of the first point */
  double q2[2];	/* IN: position (ra,dec) of the second point*/
{
  double da;
    da = q2[0] - q1[0];
    if (da < 0.e0) da += 360.e0;
    return (da * DEG * fabs( sind(q2[1]) - sind(q1[1]) ) );
}

/*===========================================================================*/
