/*++++++++++++++
.IDENTIFICATION tranum.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT
.KEYWORDS       Interpret an ascii string into numeric values
.VERSION   1.0	03-Feb-1987: Creation
.VERSION   1.1	13-Oct-1988: Cosmetic Modifications. Error tracing changed
.VERSION   2.0	18-Feb-1992: Tested with TurboC
.VERSION   3.0  10-Nov-1993: Merged tral / traf / traq
.VERSION   3.1  03-Mar-1995: Accept blanks after sign for Sexagesimal
.VERSION   3.2  11-Apr-1995: Accept RA either decimal degrees or HH MM
.VERSION   3.3  16-Apr-1996: Accept ':' in sexagesimal
				(function tr_aq)
.VERSION   3.4  06-May-1998: Decimal ONLY if a decimal point present...
.VERSION   3.5  26-Aug-1998: Accept blank if only 1 blank present.
.VERSION   3.6  17-Apr-2002: Accept exponential notation also in degrees.
.VERSION   3.7  13-Sep-2003: RA outside range [0,360[ ==> ERROR !!
			Accept also RA Dec without Sign
.VERSION   3.8  18-Sep-2008: Accept unicode extensions (3-byte multi-part)
.VERSION   3.9  29-Sep-2008: safe 64-bit
.VERSION   3.91 21-Sep-2009: Bug
.VERSION   4.0  02-Jul-2010: Function tr_error()
.VERSION   4.01 28-Aug-2010: 64-bit
.VERSION   4.1  07-Jan-2013: Accept just a blank between RA/Dec
.VERSION   4.11 20-Mar-2013: Accept a comma
.VERSION   4.12 21-May-2013: Accept a 'h' for hours
.COMMENTS       Note: NO BLANK normally can follow the sign!
---------------*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef int4
#define int4 int
#define int8 long long
#endif


/*===========================================================================
	Static variables for communications of internal routines
 *===========================================================================*/

static struct {			/* Error report	*/
  char *str;	/* Text where error occured	*/
  char *end;	/* Indicates end of str		*/
 } Str;	/* Error Structure */

#if 0
static	double 	faccu;
static  int4   *pnul = (int4 *)(&faccu);
#endif
static  int	base = 10;
static  int	ndec;			/* Number of decimals 		*/
static  char   *stopped;		/* Tell where scan stopped 	*/
static	char	sign;
static	char	err_level     = 1;	/* 0 for Warning    		*/
static  char	gobble;	/* Set to 1 if Sign accepts blanks after sign 	*/
/*atic	char	sexanum;		-- Set by ag0 if sexagesimal 	*/

static double 	pow10[] = {1.0, 10., 100., 1.e3, 1.e4, 1.e5, 1.e6, 1.e7,
	1.e8, 1.e9};

static char amsg[256];
static FILE *error_file;	/* Default = stderr, changed by tr_error */

/*===========================================================================
		Error report
 *===========================================================================*/

static int error(char *msg, char *err)
/*++++++
.PURPOSE Display Error.
.RETURNS 0 (no error) / 1 (error)
--------*/
  /* char *msg :IN: Error Message */
  /* char *err :IN: Position of bad character */
{
  char *p, *e;
    if (! msg)	return(0);
#if 0
    if (display_error) {
    	c = *(Str.end), *(Str.end) = 0;
  	fprintf(stderr, "%s%s in: %s\n         ",
  	    err_level ? "****" : "++++", msg, Str.str);
  	for (p = msg; *p; p++) fputc(' ', stderr);
  	for (p = Str.str; p < err; p++) fputc(' ', stderr);
  	fputc('^', stderr); fputc('\n', stderr);
    	*(Str.end) = c;
    }
#else
    if (!error_file) error_file = stderr;
    p = amsg; e = p + sizeof(amsg) -1;
    strcpy(p, err_level ? "****" : "++++"); p += strlen(p);
    strncpy(p, msg, (e-p)); p += strlen(p);
    if ((e-p)>24) {
	/* Specify the offset if no output file */
	strcpy(p, " in: ");
	if (error_file == (FILE *)(-1))
	    sprintf(p, " in[%d]: ", (int)(err - Str.str));
	p += strlen(p);
	strncpy(p, Str.str, (e-p));
    }
    if (error_file != (FILE *)(-1))
        fprintf(error_file, "%s\n", amsg);
#endif
    return(1);
}

static int warning(char *msg, char *err)
/*++++++
.PURPOSE Display Warning message
.RETURNS 0 (no error) / 1 (error)
--------*/
  /* char *msg :IN: Error Message */
  /* char *err :IN: Position of bad character */
{
    err_level = 0;
    error(msg, err);
    err_level = 1;
    return(0);
}

/*===========================================================================
		Internal routines for interpretation of bytes
 *===========================================================================*/

/* Note on V3.8: Unicode characters aaaa bbbb bbcc cccc are coded on 3 bytes:
 * 1110 aaaa  10bb bbbb  10cc cccc  i.e.  Unicode x2212 (mth. minus) becomes
 * 1110 0010  1000 1000  1001 0010   or   E2 88 92
 */
#define isblank(c)		(((c)==' ')||((c)=='\xa0'))  /* V3.8: nbsp */
#define isminus(c)		(((c)=='-')||((c)=='\x96')||((c)=='\x97'))

#define MAX(a,b)		((a) >= (b) ? (a) : (b))
#define MIN(a,b)		((a) <= (b) ? (a) : (b))
#define int4NULL		0x80000000		/* if inexistant */
#define int8NULL		0x8000000000000000LL	/* if inexistant */
#define GobbleSpaces(s,e)	while ((s < e) && (isblank(*s))) s++
#define issign(c)		(((c) == '+') || isminus(c))
				/* (((c) == '+') || ((c) == '-')) */



#if 1
#define isminus3(a)		(((a)[0]=='\xe2')&&( \
	(((a)[1]=='\x88')&&((a)[2]=='\x92')) ||	   /* Unicode \u2212 */ \
	(((a)[1]=='\x80')&&(((a)[2]&0xfc)==0x90))  /* Unicode \u2010..2013 */\
	))
#else	/* Implement as a function */
static int isminus3(char *a) {
    if (a[0] != '\xe2') return(0);
    if ((a[1] == '\x88') && (a[2]=='\x92')) return(1);
    if ((a[1] == '\x80') &&((a[2]&0xfc)==0x90)) return(1);
    return(0);
}
#endif
/*===========================================================================*/
static int Number(char *str, int len, int4 *accu)
/*++++++
.PURPOSE Just scan a string until non-valid characters are found.
	The number is stored in accu.
.RETURNS Number of valid characters
--------*/
  /* char *str :IN: The string to scan	*/
  /* int  len  :IN: Length of the string	*/
{
  register char *p, *e;
  register int  i;

    *accu = 0;
    if (len <= 0)		return(0);

    for (p = str, e = p + len; (p < e) && (isxdigit(*p)); p++) {
  	if (isdigit(*p))	i = '0';
  	else if (isupper(*p))	i = 'A' - 10;
	else			i = 'a' - 10;
	i = *p - i;
	if (i >= base)		break;
	*accu = *accu * base + i;
    }
    i = p - str;		/* Number of Digits */

    return(i);
}

static int Num8(char *str, int len, int8 *accu)
/*++++++
.PURPOSE Just scan a string until non-valid characters are found.
	The number is stored in accu.
.RETURNS Number of valid characters
--------*/
  /* char *str :IN: The string to scan	*/
  /* int  len  :IN: Length of the string	*/
{
  register char *p, *e;
  register int  i;

    *accu = 0;
    if (len <= 0)		return(0);

    for (p = str, e = p + len; (p < e) && (isxdigit(*p)); p++) {
  	if (isdigit(*p))	i = '0';
  	else if (isupper(*p))	i = 'A' - 10;
	else			i = 'a' - 10;
	i = *p - i;
	if (i >= base)		break;
	*accu = *accu * base + i;
    }
    i = p - str;		/* Number of Digits */

    return(i);
}

/*===========================================================================*/
static int Sign(char *str, int len)
/*++++++
.PURPOSE Just scan the beginning of a string: skip leading blanks,
	and get the sign.
.RETURNS Number of valid characters
--------*/
  /* char *str :IN: The string to scan	*/
  /* int  len :IN: Length of the string	*/
{
  char *p, *e;

    sign = 0;
    if (len <= 0)		return(0);

    p = str, e = p + len;
    GobbleSpaces(p, e);

    if (p < e) {
	if (*p == '+')           p++;
	else if (isminus(*p)) { sign = 1; p++; }
	else if (isminus3(p)) { sign = 1; p += 3; }
	if (gobble) GobbleSpaces(p, e);		/* Skip blanks */
    }
    return(p-str);
}

/*===========================================================================*/
static int traf(char *str, int len, double *faccu)
/*++++++
.PURPOSE Converts a string to a floating point.
	Input is assumed to be in decimal.
.RETURNS Number of significant digits
.REMARKS Return 0 if string contains only spaces.
	A static ndec contains the number of digits following the .
	Void double is set to (ff)
--------*/
  /* char  *str :IN: String to scan	*/
  /* int    len :IN: Length of str	*/
{
  char 	*p, *e, *pexp, x, xexp=0;
  int	nd;
  int4  *pnul;
  double atof();

#ifdef DEBUG
    printf("....traf: examine %s\n", str);
#endif

    nd = ndec = 0;
    pnul = (int4 *)faccu;
    pnul[0] = pnul[1] = 0xffffffff; 	/* NaN */
    stopped = str;
    p = str + Sign(str, len);	/* Skip header + sign	*/
    e = str + len;
    if (p == e)	return(0);	/* Empty string */

    pexp = (char *)0;
    			/* Skip leading zeroes, not significant */
    while ((p<e) && (*p == '0')) { *faccu = 0; p++; }
    stopped = p;
    if (p == e)	return(0);		/* Zero	*/

    while ((p<e) && isdigit(*p)) p++, nd++;
    if (p == e)	goto get_value;
    if (*p == '.')	p++;
    if (nd == 0)			/* Skip leading zeroes	*/
    	while ((p<e) && (*p == '0')) p++, ndec++;
    while ((p<e) && isdigit(*p)) p++, nd++, ndec++;

	/* Look for Exponent */

    if ((p < e) && (isalpha(*p))) {
    	pexp = p;
  	xexp = toupper(*p);
  	if ((xexp == 'E') || (xexp == 'D')) {	/* Exponent */
  	    *pexp = 'e';
  	    p++;
  	    if    (issign(*p))	p++;
  	    while (isdigit(*p)) p++;
  	}
    }

  get_value:
    x = *p, *p = '\0', *faccu = atof(str), *p = x;
    if (pexp) *pexp = xexp;

    GobbleSpaces(p, e);		/* Skip trailing blanks */
    stopped = p;
    return(nd);
}

/*===========================================================================*/
static int ag0(char *str, int len, double *faccu, int *sexa)
/*++++++
.PURPOSE Converts a string to a double value; input may be
	in sexagesimal.
.RETURNS The precision: -1=error, 0=void, 1=deg, 2=0.1deg,
		3=1', 4=0.1', 5=1", 6=0.1", 7=0.01", ...
.REMARKS Keeps the position where the scan stopped.
	sexa contains in return a non-zero with
	1/2 ==> there are 1/2 components
   ORed with 4 => separator is colon
--------*/
  /* char *str :IN: String to scan	*/
  /* int  len  :IN: Length of str	*/
{
  register	char 	*p, *e, sep;
  register	int	nd, prec, sexanum;
  int4 		accu;
  double	factor;

    nd = gobble; 		/* Save global variable */
    gobble = 1;			/* Tell Sign we accept blanks after sign */
    p = str + Sign(str, len);	/* Skip header + sign	*/
    gobble = nd;		/* Restore 		*/
    e = str + len;
    base   = 10;
    if ((p[0] == ' ') && isdigit(p[1])) p++;

	/* Convert the degrees number	*/
    sexanum = 1;
    factor = 1.0;
    sep    = 0 ;	/* Separator h:m, 0 or ':'	*/
    nd     = Number(p, e-p, &accu);  p += nd;
    *faccu = accu;
    prec   = nd != 0;	/* lead to 1 for degrees */
    if (p >= e)		goto terminate;
    if (*p == '.')	{ sexanum = 0 ; goto get_decimals; }

	/* Convert the minutes		*/
    if (*p == ':') sep = ':', p++ ;
    else GobbleSpaces(p, e);
    if (issign(*p) || ((*p&0xfc)==0xfc)) /* V3.8: Unicode */
	goto terminate;
    factor *= 60.0;
    nd      = Number(p, e-p, &accu);
    if (accu >= 60)	warning("minutes above 60", p);
    p      += nd;
    *faccu += accu/factor;
    if (nd) prec += 2;
    if (p >= e)		goto terminate;
    if (*p == '.')	goto get_decimals;

	/* Convert the seconds		*/
    if (sep) {
	if (*p != sep) goto terminate;
	p++ ;
    }
    GobbleSpaces(p, e);
    factor *= 60.0;
    nd      = Number(p, e-p, &accu);
    if (accu >= 60)	warning("seconds above 60", p);
    if (nd>0) sexanum = 2;	/* There are 3 components */
    p      += nd;
    *faccu += accu/factor;
    if (nd) prec += 2;
    if (p >= e)		goto terminate;

	/* Convert the decimals		*/
  get_decimals:
    if (*p == '.') {
    	p++;
    	while (p < e) {
    	    if (! isdigit(*p)) break;
    	    nd = e-p;
    	    if (nd > 9) nd = 9;		/* To avoid long overflow */
    	    nd = Number(p, nd, &accu); p += nd;
    	    prec   += nd;
    	    factor *= pow10[nd];
    	    *faccu += accu/factor;
    	}
	/* Addition V3.6: Accept with Exponent */
	if ((p<e) && (sexanum == 0) && (tolower(*p) == 'e')) {
	    p++ ;
	    sep = 0 ;	/* Used here for the sign of the exponent */
	    if (issign(*p)) sep = *(p++) ;
	    if (isminus(sep)) prec += nd; else prec -= nd;
	    for (nd=0; (p<e) && isdigit(*p); p++)
	       nd = nd*10 + (*p - '0') ;
	    for (factor = 1.; nd>9; nd -= 9) factor *= pow10[9];
	    factor *= pow10[nd];
	    if (isminus(sep)) *faccu /= factor ;
	    else *faccu *= factor ;
	}
    }

  terminate:
    GobbleSpaces(p, e);
    if (sign)	*faccu = -*faccu;
    if (sexa) { *sexa = sexanum; if(sep) *sexa |= 4; }
    stopped = p;

    return(prec);
}


/*===========================================================================
		Dealing with Long integers
 *===========================================================================*/

char *tr_error(FILE *errfile)
/*++++++
.PURPOSE Specifies the file where the error should be printed, and
	returns the last error string
.RETURNS Last error message (internal buffer) / empty string
--------*/
{
    error_file = errfile ? errfile : (FILE *)(-1);
    return(amsg);
}

int tr_al(char *str, int len, int4 *value)
/*++++++
.PURPOSE Converts a string to a long number.
	Input is assumed to be in decimal, unless the string starts with
	' (quote)  ==> character
	^ (carret) ==> control-character
	X          ==> hexadecimal or
	O 	   ==> octal.
	\	   ==> C-char escape sequence
	The digits may be followed by K (implies a 1024 factor)
.RETURNS Number of digits (excluding spaces) scanned, -1 for
	error (message logged)
.REMARKS Return 0 if string contains only spaces or a sign.
	Value is then set to zero.
--------*/
  /* char  *str    :IN: String to scan	*/
  /* int   len     :IN: Length of str	*/
  /* int4  *value :OUT: result		*/
{
  char 	*p, *e, *err;
  int4	nd, accu;

    err = (char *)0;

    base = 10;
    nd = 1;
    accu = 0;

    if (len<0) len = strlen(str);
    Str.str = p = str + Sign(str, len);	/* Skip header + sign	*/
    Str.end = e = str + len;

    if (*p == '\'') {		/* Get a Character */
    	p++;
	if (*p != '\'')	accu = *(p++);
	if (*p == '\'')	p++;
	goto check_remaining_bytes;
    }

    if (*p == '\\') {		/* Get a C-string sequence */
    	p++;
    	switch(tolower(*p)) {
  	  case 'a': accu =  007; break;	/* Bell   */
  	  case 'b': accu =  010; break;	/* Backsp */
  	  case 'e': accu =  033; break;	/* Escape */
  	  case 'f': accu = '\f'; break;
  	  case 'n': accu = '\n'; break;
  	  case 'r': accu = '\r'; break;
  	  case 't': accu = '\t'; break;
  	  default : p--;	/* Generates error */
  	}
  	p++;
	goto check_remaining_bytes;
    }

    if (*p == '^') {		/* Get a Control Character */
    	p++;
  	accu = *p == '?' ? 0177 : *p & 037;
  	p++;
	goto check_remaining_bytes;
    }

    nd = Number(p, e-p, &accu);	/* Get number		*/
    p += nd;			/* End of number	*/

	/* If the string is not completely understood, look if
	   it can be due to O(octal), X(hexadecimal), k (k-factor)
    	   Note that X and O can only be at the beginning, i.e.
    	   when accu == 0; k can only be a suffix.
	 */

    if (p < e)	switch(toupper(*p)) {
      case 'O':	base = 8;
  	if (accu == 0)	goto special_base;
  	break;
      case 'X':	base = 16;
  	if (accu == 0)	goto special_base;
	break;
      special_base:
	p++;
	nd = Number(p, e-p, &accu);
	p += nd;
	break;
      case 'K':
    	accu <<= 10;
    	p++;
	break;
    }

	/* Take Sign into account, and check remaining bytes are only
	   white spaces
	 */

    check_remaining_bytes:
    *value = (sign ? -accu : accu);

    GobbleSpaces(p, e);

    if (p < e) {		/* Bad String	*/
  	switch(base) {
  	  case 8:	err = "Bad octal number";	break;
  	  case 10:	err = "Invalid digit";		break;
  	  case 16:	err = "Invalid hexa-digit";	break;
  	}
	nd = -1;
    }
    if (err) error(err, p);

    return(nd);
}

int tr_all(char *str, int len, int8 *value)
/*++++++
.PURPOSE Converts a string to a 64-bit number.
	Input is assumed to be in decimal, unless the string starts with
	X          ==> hexadecimal or
	O 	   ==> octal.
.RETURNS Number of digits (excluding spaces) scanned, -1 for
	error (message logged)
.REMARKS Return 0 if string contains only spaces or a sign.
	Value is then set to zero.
--------*/
  /* char  *str    :IN: String to scan	*/
  /* int   len     :IN: Length of str	*/
  /* int8  *value :OUT: result		*/
{
  char 	*p, *e, *err;
  int4	nd;

    err = (char *)0;

    base = 10;
    nd = 1;
    *value = 0;

    if (len<0) len = strlen(str);
    Str.str = p = str + Sign(str, len);	/* Skip header + sign	*/
    Str.end = e = str + len;

    nd = Num8(p, e-p, value);	/* Get number		*/
    p += nd;			/* End of number	*/

	/* If the string is not completely understood, look if
	   it can be due to O(octal), X(hexadecimal), k (k-factor)
    	   Note that X and O can only be at the beginning, i.e.
    	   when value == 0; k can only be a suffix.
	 */

    if (p < e)	switch(toupper(*p)) {
      case 'O':	base = 8;
  	if (*value == 0)	goto special_base;
  	break;
      case 'X':	base = 16;
  	if (*value == 0)	goto special_base;
	break;
      special_base:
	p++;
	nd = Num8(p, e-p, value);
	p += nd;
	break;
    }

	/* Take Sign into account, and check remaining bytes are only
	   white spaces
	 */

    if (sign) *value = -*value;

    GobbleSpaces(p, e);

    if (p < e) {		/* Bad String	*/
  	switch(base) {
  	  case 8:	err = "Bad octal number";	break;
  	  case 10:	err = "Invalid digit";		break;
  	  case 16:	err = "Invalid hexa-digit";	break;
  	}
	nd = -1;
    }
    if (err) error(err, p);

    return(nd);
}


/*===========================================================================
		Interpret array of numbers
 *===========================================================================*/

int tr_al1(char *str, int len, int4 *value, int n)
/*++++++
.PURPOSE Converts a string to a vector of INTEGER numbers.
	Numbers may be separated by blanks, commas,  or any punctuation character,
	but can only be in decimal
.RETURNS Number of found numbers (<= n) / -1 for error (message logged)
.REMARKS Return 0 if string contains only spaces.
	Values with only blanks or a sign are set to the smallest number
	(0x80000000)
--------*/
  /* char *str    :IN: String to scan	*/
  /* int   len    :IN: Length of str	*/
  /* int4 *value :OUT: array result	*/
  /* int   n      :IN: Number of values	*/
{
  char 	*p, *e, *err, ebuf[40];
  int	i, nd, accu;

    base = 10;
    err = (char *)0;
    if (len<0) len = strlen(str);
    Str.str = p = str;
    Str.end = e = str + len;

    for (i=0; (p<e) && (i<n); i++) {
  	p += Sign(p, e-p);
  	if (p == e)	break;
	nd = Number(p, e-p, &accu);
	if (nd < 0) {
  	    err = "Invalid digit";
  	    break;
	}
  	if (nd > 0) value[i] = (sign ? -accu : accu);
  	p += nd;
  	GobbleSpaces(p, e);
	if (issign(*p))	continue;
	if (ispunct(*p)) p++;
    }

    /* Check string exhausted */
    while ((p<e)&&isspace(*p)) p++;
    if (i<n) value[i] = int4NULL;

    if ((!err) && (p < e))
    	err = ebuf, sprintf(ebuf, "Too many (>%d) numbers", n);
    if (err)
    	i = -1, error(err, p);

    return(i);
}

int tr_all1(char *str, int len, int8 *value, int n)
/*++++++
.PURPOSE Converts a string to a vector of INTEGER numbers.
	Numbers may be separated by blanks, commas,  or any punctuation character,
	but can only be in decimal
.RETURNS Number of found numbers (<= n) / -1 for error (message logged)
.REMARKS Return 0 if string contains only spaces.
	Values with only blanks or a sign are set to the smallest number
	(0x8000000000000000)
--------*/
  /* char *str    :IN: String to scan	*/
  /* int   len    :IN: Length of str	*/
  /* int4 *value :OUT: array result	*/
  /* int   n      :IN: Number of values	*/
{
  char 	*p, *e, *err, ebuf[40];
  int	i, nd;
  int8  accu;

    base = 10;
    err = (char *)0;
    if (len<0) len = strlen(str);
    Str.str = p = str;
    Str.end = e = str + len;

    for (i=0; (p<e) && (i<n); i++) {
  	p += Sign(p, e-p);
  	if (p == e)	break;
	nd = Num8(p, e-p, &accu);
	if (nd < 0) {
  	    err = "Invalid digit";
  	    break;
	}
  	if (nd > 0) value[i] = (sign ? -accu : accu);
  	p += nd;
  	GobbleSpaces(p, e);
	if (issign(*p))	continue;
	if (ispunct(*p)) p++;
    }

    /* Check string exhausted */
    while ((p<e)&&isspace(*p)) p++;
    if (i<n) value[i] = int8NULL;

    if ((!err) && (p < e))
    	err = ebuf, sprintf(ebuf, "Too many (>%d) numbers", n);
    if (err)
    	i = -1, error(err, p);

    return(i);
}

/*===========================================================================*/
int tr_aln(char *str, int len, int4 *value, int n)
/*++++++
.PURPOSE Converts a string to a vector of numbers.
	Numbers may be separated by blanks or any punctuation character;
	there must be exactly the required number of values.
.RETURNS  n for success / -1 for failure
.REMARKS Return 0 if string contains only spaces or a sign.
	Values are then set to minimal values (0x80000000).
--------*/
  /* char *str    :IN: String to scan	*/
  /* int   len    :IN: Length of str	*/
  /* int4 *value :OUT: array result	*/
  /* int   n      :IN: Number of values	*/
{
  int	i;
  char ebuf[60];

    if (len<0) len = strlen(str);
    i = tr_al1(str, len, value, n);

    if ((i >= 0) && (i < n)) {	/* Too few numbers */
  	sprintf(ebuf, "Too few (%d < %d) numbers", i, n);
  	error(ebuf, str+len);
  	i = -1;
    }

    return(i);
}

int tr_alln(char *str, int len, int8 *value, int n)
/*++++++
.PURPOSE Converts a string to a vector of numbers.
	Numbers may be separated by blanks or any punctuation character;
	there must be exactly the required number of values.
.RETURNS  n for success / -1 for failure
.REMARKS Return 0 if string contains only spaces or a sign.
	Values are then set to minimal values (0x80000000).
--------*/
  /* char *str    :IN: String to scan	*/
  /* int   len    :IN: Length of str	*/
  /* int4 *value :OUT: array result	*/
  /* int   n      :IN: Number of values	*/
{
  int	i;
  char ebuf[60];

    if (len<0) len = strlen(str);
    i = tr_all1(str, len, value, n);

    if ((i >= 0) && (i < n)) {	/* Too few numbers */
  	sprintf(ebuf, "Too few (%d < %d) numbers", i, n);
  	error(ebuf, str+len);
  	i = -1;
    }

    return(i);
}

/*===========================================================================
		Getting floating-point number
 *===========================================================================*/

int tr_af(char *str, int len, double *value)
/*++++++
.PURPOSE Converts a string to a floating point.
	Input is assumed to be in decimal.
.RETURNS Number of significant digits / -1 for
	error (message logged)
.REMARKS Return 0 if string contains only spaces.
--------*/
  /* char  *str     :IN: String to scan	*/
  /* int    len     :IN: Length of str	*/
  /* double *value :OUT: result	*/
{
  char 	*p, *e;
  int	nd;

    nd = 0;
    if (len<0) len = strlen(str);
    Str.str = p = str;
    Str.end = e = str + len;
    nd = traf(p, len, value);

    p = stopped;
    GobbleSpaces(p, e);

    if (p != e) {
  	error("Invalid digit", p);
	nd = -1;
    }

    if (nd == 0) { int4 *pnul = (int4 *)value;
	if (pnul[0] != 0xffffffff) nd = 1;
    }
    return(nd);
}

/*===========================================================================*/
int tr_ao0(char *str, int len, double *g0)
/*++++++
.PURPOSE Converts a string to a double value; input may be
	in sexagesimal.
.RETURNS The precision, from 0 (no data), 1 (degree) to 8 (.001"),
	or -1 (error)
.REMARKS If an error is encountered, message logged.
	NaN is returned when no coordinate exist (0 returned).
--------*/
  /* char  *str   :IN: String to scan	*/
  /* int    len   :IN: Length of str	*/
  /* double  *g0 :OUT: result		*/
{
  int nd;
    if (len<0) len = strlen(str);
    Str.str = str;
    Str.end = str + len;

    nd = ag0(Str.str, len, g0, 0);
    if (stopped != Str.end) {
	nd = -1;
	error ("Bad sexagesimal number:", stopped);
    }
    if (nd <= 0)  {
	int4 *pnul = (int4 *)g0;
	pnul[0] = pnul[1] = 0xffffffff; 	/* NaN */
    }

    return(nd);
}

/*===========================================================================*/
int tr_af1(char *str, int len, double *value, int n)
/*++++++
.PURPOSE Converts a string to a vector of numbers.
	Numbers may be separated by blanks or any punctuation character,
.RETURNS Number of found numbers (<= n) / -1 for
	error (message logged)
.REMARKS Return 0 if string contains only spaces.
--------*/
  /* char  *str     :IN: String to scan	*/
  /* int    len     :IN: Length of str	*/
  /* double *value :OUT: result	*/
  /* int    n       :IN: Size of values	*/
{
  char 	*p, *e, ebuf[60];
  int	i;

    if (len<0) len = strlen(str);
    Str.str = p = str;
    Str.end = e = str + len;

    for (i=0; (p<e) && (i < n); i++) {
  	GobbleSpaces(p, e);
	if (p == e)	break;
	if (traf(p, e-p, value+i) < 0) return(-1);
	p = stopped;
	if (p == e)		continue;
	if (issign(*p))		continue;
	if (isdigit(*p))	continue;
	if (*p == '.')		continue;
	p++;
    }
    /*if (i<n) value[i] = -0./0.; *** DON't MODIFY VALUES NOT READ !! */

    /* Check string exhausted */
    while ((p<e)&&isspace(*p)) p++;

    if (p != e) {
    	sprintf(ebuf, "Too many (>%d) numbers", n);
    	error(ebuf, p);
    	i = -1;
    }
    return(i);
}

/*===========================================================================*/
int tr_afn(char *str, int len, double *value, int n)
/*++++++
.PURPOSE Converts a string to a vector of double numbers.
	Numbers may be separated by blanks or any punctuation character.
.RETURNS  n for success / -1 for failure
.REMARKS
--------*/
  /* char  *str     :IN: String to scan	*/
  /* int    len     :IN: Length of str	*/
  /* double *value :OUT: array result	*/
  /* int    n       :IN: Number of values	*/
{
  char ebuf[60];
  int	i;

    i = tr_af1(str, len, value, n);

    if ((i >= 0) && (i < n)) {	/* Too few numbers */
  	sprintf(ebuf, "Too few (%d < %d) numbers", i, n);
  	error(ebuf, str+len);
  	i = -1;
    }

    return(i);
}

/*===========================================================================
		Dealing with Sexagesimal Coordinates
 *===========================================================================*/

static int tr_aoq(char *str, int len, double *coo, int flag)
/*++++++
.PURPOSE Converts a string to (longitude, latitude)
.RETURNS The accuracy, from -1(error), 0(no data), 1(deg)... 7 (0.01"), 8(mas)..
.REMARKS Longitude in range [0,360[. Not traced.
--------*/
  /* char  *str     :IN: String to scan	*/
  /* int    len     :IN: Length of str	*/
  /* double coo[2] :OUT: resulting coordinates		*/
  /* int    flag    :IN: Flag 1 for RA + DEC, 0 for long + lat	*/
{
  register char  *p;
  register int	ndl, ndb, hr=0;
  int4 sexanum, *pnul;

    Str.str = str;
    Str.end = str + len;

	/* Get Longitude	*/
    ndl = ag0(Str.str, len, coo, &sexanum);   /* Get Longitude */
    ndb = 0 ;			/* Used as a flag to indicate an error */
#if 0	/* Change in v3.7 */
    if (*stopped == ',') {	/* Accept the comma for the HEASARC syndrome */
	stopped++ ;
	GobbleSpaces(stopped, Str.end) ;
    }
    else if (!issign(*stopped)){/* Accept blank if only 1 blank present */
	GobbleSpaces(stopped, Str.end) ;
	for (p=stopped; (p<Str.end) && (*p != ' '); p++) ;
	ndb = Str.end - p ;	/* Not zero if blank found, i.e. error	*/
    }
#else	/* Change in v3.7 */
    if (stopped < Str.end) {
        if (*stopped == 'h')  { 	/* V4.12: accept 'h' */
	    hr=1; stopped++; 
	    if (*stopped == 'r') stopped++;	/* Accept "hr" */
        }
        if ((*stopped == ',') || (*stopped == '\t')) {
	    /*sexanum |= 8;	*//* There is a separator */
	    stopped++;			/* Accept the comma for HEASARC	*/
        }
        GobbleSpaces(stopped, Str.end);
    }
#if 0
    if (isdigit(*stopped));		/* Accept without sign in Dec	*/
    else if (issign(*stopped)) ;
    else if (!isminus3(stopped))
	ndb = -1;			/* ERROR */
#else
    /* V4.1: accept without sign in Dec */
    if (stopped == Str.end){
	if ((sexanum>=1) && (sexanum<=2)) {
	    /* The last part is the Declination */
	    for (--stopped; isblank(*stopped); --stopped);
	    while (isdigit(*stopped)||(*stopped=='.')) stopped--;
	    ++stopped;
	    ndl = ag0(Str.str, stopped-Str.str, coo, &sexanum);
	}
	else ndb = -1;			/* ERROR */
    }
    else if (isdigit(*stopped));	/* Accept without sign in Dec	*/
    else if (issign(*stopped)) ;
    else if (!isminus3(stopped))
	ndb = -1;			/* ERROR */
#endif
#endif
    if (ndb) {	/* V4.1 */
	ndl = -1;
	error ("Missing sign between RA(lon) and Dec(lat)", stopped);
    }
    if (ndl <= 0)  { 	/* NaN */
	pnul = (int4 *)coo; 
	pnul[0] = pnul[1] = 0xffffffff; 
    }
    if (ndl < 0) {	/* Error => NaN for Dec */
    	coo[1] = coo[0];
    	return(ndl);
    }

	/* Convert hours to degrees --- but only if minutes are present	*/
#if 0
    if (flag && sexanum) {
  	coo[0] *= 15.e0;		/* Convert to degrees	*/
	if (ndl > 1)	ndl--;
    }
#else	/* V4.12: accept hours */
    if (hr || (flag && sexanum)) {
  	coo[0] *= 15.e0;		/* Convert to degrees	*/
	if (ndl > 1)	ndl--;
    }
#endif

	/* Longitude in range [0,360]	*/
#if 0
    while (coo[0] <    0.0)	coo[0] += 360.0;
    while (coo[0] >= 360.0)	coo[0] -= 360.0;
#else
    if (coo[0] < 0) coo[0] += 360.0;
    if ((coo[0] < 0) || (coo[0] >= 360.)) {
	ndl = -1;
	error ("RA(lon) outside range [0, 360[", str);
    }
#endif
	/* Get Latitude		*/
    p   = stopped;
    ndb = ag0(p, Str.end - p, coo+1, 0);
    if ((coo[1] < -90.e0) || (coo[1] > 90.e0)) {
	ndb = -1;
	error ("Dec(lat) outside range [-90,+90]", p);
    }

    /* Accept more spaces, \n, \r, etc... (mod. V3.91) */
    while ((stopped<Str.end) && isspace(*stopped)) stopped++;

    if (stopped != Str.end) {
	ndb = -1;
	error ("Bad Dec(lat):", stopped);
    }
    if (ndb <= 0)  { /* NaN */
	pnul = (int4 *)(&coo[1]);
	pnul[0] = pnul[1] = 0xffffffff; 	/* NaN */
    }

	/* Final Precision	*/
    return(MIN (ndl, ndb));
}

/*===========================================================================*/
int tr_ao(char *str, int len, double *coo)
/*++++++
.PURPOSE Converts a string to (longitude, latitude)
.RETURNS The accuracy, from -1 (error), 0(no data), 1(degrees), 2(0.1 deg),
	... 5(1"), 6(0.1"), ...
.REMARKS Longitude in range [0,360[.
--------*/
  /* char  *str     :IN: String to scan	*/
  /* int   len      :IN: Length of str	*/
  /* double coo[2] :OUT: resulting coordinates	*/
{
  return(tr_aoq(str, len, coo, 0));
}

/*===========================================================================*/
int tr_aq(char *str, int len, double *coo)
/*++++++
.PURPOSE Converts a string to (right ascension, declination)
.RETURNS The accuracy, from -1 (error), 0(no data), 1(degrees), 2(0.1 deg),
	... 5(1"), 6(0.1"), ...
.REMARKS Right ascension in degrees, in range [0,360[
--------*/
  /* char  *str     :IN: String to scan	*/
  /* int    len     :IN: Length of str	*/
  /* double coo[2] :OUT: resulting coordinates	*/
{
  return(tr_aoq(str, len, coo, 1));
}

/*===========================================================================
		Main program (#if TEST)
 *===========================================================================*/

#ifdef TEST
#define ASIZE 12

int main()
/*++++++++++++++++
.PURPOSE  TEST PROGRAM
.RETURNS  to OS
-----------------*/
{
  char msg[133];
  int4 	 lval[ASIZE];
  double fval[ASIZE];
  int  n, i, asize, choice;

    puts("Testing char -> Binary\n");
    for(;;) {
  	printf("----Single (S), Double (F), Integer (I), coordinates (C) ? ");
    	if(!fgets(msg, sizeof(msg), stdin)) break;
    	switch(toupper(msg[0])) {
    	  case 'F': case 'E': case 'R':
    	    choice = 4; break;
    	  case 'C':
    	    choice = 2;
    	    break;
    	  case 'I':
    	    choice = 0;
    	    break;
      	  default :
      	    puts("****I can't understand (sorry)");
      	    continue;
    	}

	if (choice != 2) {
    	    printf("----Size of the array (<=1 for single) ");
    	    fgets(msg, sizeof(msg), stdin);
	    asize = atoi(msg);
	    if (asize <= 1) asize = 0;
	    if (asize > ASIZE) {
	        printf("++++Size restricted to %d elements\n", ASIZE);
	        asize = ASIZE;
	    }
    	}

    	for(;;) {
    	    printf("----Your input: ");
	    if (!fgets(msg, sizeof(msg), stdin)) break;
	    for (i=strlen(msg)-1; (i>0)&&iscntrl(msg[i]); i--) msg[i] = 0;
	    switch(choice) {
	      case 0:	/* Single Integer */
	  	if (asize) n = tr_al1(msg, strlen(msg), lval, asize);
	  	else  n = tr_al(msg, strlen(msg), lval);
	  	printf("====Returned status %d\n", n);
	  	printf("\t%d", lval[0]);
	  	for (i=1; i < asize; i++) printf("\t%d", lval[i]);
	  	puts("");
	  	break;
	      case 2:	/* Coordinates */
	  	n = tr_aq(msg, strlen(msg), fval);
	  	printf("====Returned status %d\n", n);
	  	printf("\t%.16lg %.16lg\n", fval[0], fval[1]);
	  	break;
	      case 4:	/* Single Double */
	  	if (asize) n = tr_af1(msg, strlen(msg), fval, asize);
	  	else  n = tr_af(msg, strlen(msg), fval);
	  	printf("====Returned status %d\n", n);
	  	printf("\t%lg", fval[0]);
	  	for (i=1; i < asize; i++) printf("\t%lg", fval[i]);
	  	puts("");
	    }
    	}
    }
    exit(0);
}
#endif
