/*++++++++++
.TYPE           Module
.IDENTIFICATION trsp.c
.AUTHOR         Francois Ochsenbein [ESO-IPG]  
.VERSION 1.0	02-Feb-1989: Creation
.VERSION 1.1	11-May-1990: Cast for signed/unsigned char.
		Fixed (Symptom = the : for uncertainty in LC not recognized)
.VERSION 1.2	22-Jun-1990: 3 problems fixed:
			O0 / O1 / O2 did not result into an error
			K5/M0 was translated into K5/M0:...
			M10   was translated into M1...
.VERSION 1.3	13-Nov-1992: Corrected the following errors:
			sdO: was not coded correcly (lowercase luminosities)
			Iab-b was not coded correctly (-b misunderstood)
.VERSION 1.4	08-Nov-1994: Cosmetics - Stand-alone program
.VERSION 1.41	14-Apr-2012: Accept , in sp-type (used for S class)
.LANGUAGE       C
.KEYWORDS       Spectral Classification

.ENV            Any

.COMMENTS

\begin{TeX}
The temperature class is a number in the range 0---255 (8 bits), 
defined as follows:

$$\begin{tabular}{|rl|rl|rl|rl|rl|}\hline
004 & O3   & 058 & B6.5   &   114 & F1    &    174 & K1.5   &      &	\\
008 & O4   & 060 & B7     &   116 & F2    &    176 & K2     &      &    \\
010 & O4.5 & 062 & B7.5   &   118 & F2.5  &    178 & K2.5   &   236 & R \\
012 & O5   & 064 & B8     &   120 & F3    &    180 & K3     &   237 & N \\
014 & O5.5 & 066 & B8.5   &   122 & F4    &    182 & K3.5   &   238 & S \\
016 & O6   & 068 & B9     &   124 & F5    &    184 & K4     &   239 & C \\
018 & O6.5 & 069 & B9.2   &   126 & F5.5  &    186 & K4.5   &       &    \\
020 & O7   & 070 & B9.5   &   128 & F6    &    188 & K5     &       &    \\
022 & O7.5 & 071 & B9.7   &   130 & F6.5  &    190 & K6     &       &     \\
024 & O8   & 072 & A0     &   132 & F7    &    192 & K7     &   240 & WN \\
026 & O8.5 & 074 & A0.5   &   134 & F7.5  &    194 & K8     &   241 & WR \\
028 & O9   & 076 & A1     &   136 & F8    &    196 & M0     &   242 & WC \\
029 & O9.2 & 078 & A1.5   &   138 & F9    &    198 & M0.5   &       &     \\
030 & O9.5 & 080 & A2     &   140 & G0    &    200 & M1     &       &     \\
031 & O9.7 & 082 & A2.5   &   142 & G0.5  &    202 & M1.5   &   244 & DC \\
032 & B0   & 084 & A3     &   144 & G1    &    204 & M2     &       &     \\
033 & B0.2 & 086 & A3.5   &   146 & G1.5  &    206 & M2.5   &   246 & DO\\
034 & B0.5 & 088 & A4     &   148 & G2    &    208 & M3     &   247 & DB\\
035 & B0.7 & 090 & A4.5   &   150 & G2.5  &    210 & M3.5   &   248 & DA\\
036 & B1   & 092 & A5     &   152 & G3    &    212 & M4     &   249 & DF\\
038 & B1.5 & 094 & A5.5   &   154 & G4    &    214 & M4.5   &   250 & DG\\
040 & B2   & 096 & A6     &   156 & G5    &    216 & M5     &       &      \\
042 & B2.5 & 098 & A6.5   &   158 & G5.5  &    218 & M5.5   &       &    \\
044 & B3   & 100 & A7     &   160 & G6    &    220 & M6     &   253 & DZ\\
046 & B3.5 & 102 & A7.5   &   162 & G7    &    222 & M6.5   &   254 & DQ\\
048 & B4   & 104 & A8     &   164 & G8    &    224 & M7     &       &    \\
050 & B4.5 & 106 & A8.5   &   166 & G9    &    226 & M7.5   &       &     \\
052 & B5   & 108 & A9     &   168 & K0    &    228 & M8     &   255 & ???\\
054 & B5.5 & 110 & A9.5   &   170 & K0.5  &    232 & M9     &       &    \\
056 & B6   & 112 & F0     &   172 & K1    &    234 & M9.5   &       &    \\
\hline\end{tabular}$$

The luminosity class is a number in the range 0---31 (5 bits), 
defined as follows:

$$\begin{tabular}{|rl|rl|rl|rl|}\hline
00 & ---  & 10 & Ib-II  &   15 & IIIa    &    22 & IV-V   	\\
02 & Ia-0 & 11 & IIa    &   16 & III     &    24 & V      	\\
04 & Ia   & 12 & II     &   17 & IIIb    &    28 & VI (sd)	\\
06 & Iab  & 13 & IIb    &   18 & III-IV  &       &        	\\
08 & Ib   & 14 & II-III &   20 & IV      &       &        	\\
\hline\end{tabular}$$

The spectral classification is coded as an unsigned long integer.
It includes the folloing bits, from left to right: 
$$\begin{tabular}{ll}
Temperature class 	&	8 bits\\
Composite spectrum 	&	1 bit\\
No subclass 		&	1 bit\\
Luminosity class 	&	5 bits \\
Uncertain temperature class 	& 1 bit \\
Uncertain luminosity class 	& 1 bit \\
Incomplete spectrum (translated with \dots) &	1 bit \\
Houk's classification 	&	1 bit \\
Average (X) 		&	1 bit \\
Other source (+) 	&	1 bit \\
Quality (number) 	&	3 bits \\
Variable ({\bf v}) spectrum &	1 bit \\
emission ({\bf e}) spectrum &	1 bit \\
shell ({\bf sh}) spectrum   &	1 bit \\
peculiar ({\bf p}) spectrum &	1 bit \\
peculiar profiles: {\bf s}(1) \quad {\bf n}(2) \quad {\bf nn}(3) & 2 bits \\
peculiar elements: {\bf CN}(1) \quad {\bf w}(2) \quad {\bf m}(3) & 2 bits \\
\end{tabular}$$

\end{TeX}

-----------------------------------------------------------------------*/

#define VERSION  "1.4"
#include <spclass.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#define ITEMS(a)	sizeof(a)/sizeof(a[0])
#define MIN(a,b)	((a) < (b) ? a : b )

static unsigned char OBAFGKM[]  = {
	'O',	0x04,
	'B', 	0x20,
	'A', 	0x48,
	'F',	0x70,
	'G', 	0x8C,
	'K', 	0xA8,
	'M', 	0xC4,
	'R', 	0xEC,
	'N', 	0xED,
	'S', 	0xEE,
	'C', 	0xEF,
	'?', 	0xFF
	};
static unsigned char WR[]  = {		/* WR Stars	*/
	'N', 	0xF0,
	'R', 	0xF1,
	'C',	0xF2
	};
static unsigned char WD[]  = {		/* Degenerate Stars	*/
	'C',0xF4,
	'O',0xF6,
	'B',0xF7,
	'A',0xF8,
	'F',0xF9,
	'G',0xFA,
	'Z',0xFD,
	'Q',0xFE
	};

struct 	TYPE2 { 
	unsigned char	char1;	/* First char */
	unsigned char	no;
	unsigned char	*char2;
  };

static struct TYPE2 spectra[] = {
	{0,   7*2, &OBAFGKM[0]},			/* Normal */
	{0  , sizeof(OBAFGKM)-7*2, &OBAFGKM[7*2]},	/* 1-letter */
	{'W', sizeof(WR), &WR[0]},
	{'D', sizeof(WD), &WD[0]}
  };

/* Normal classes: list of temperature values (even indices 0,2,...)
   and a flag of 0, 1 or 5 (odd indices 1,3,...)
   Flag = 0 if type is not in the MK standards
	  5 if intermediate type exists in MK standard (O9.5, B0.5, B9.5)
	  1 for normal cases.
   If necessary, the interpolation between 2 consecutive values t0 and t1
   (corresponding to a type xx.s, e.g. G1.5) is done with the algorithm:
      t = t0 + (2 + (t1-t0)*s)/10
   Note that (t1-t0) can be 1, 2 or 4, and  1<=s<=9. 
   The resulting offset (t-t0) is:
	    s	t1-t0=1	t1-t0=2	t1-t0=4
	    ---	-------	-------	-------
	    1	    0	    0	    0
	    2	    0	    0	    1
	    3	    0	    0	    2
	    4	    0	    1	    2
	    5	    0	    1	    2
	    6	    0	    1	    3
	    7	    0	    1	    3
	    8	    1	    1	    3
	    9	    1	    2	    4
*/
static sp_class TYPEN[] = {		/* Normal Classification */
	0x00, 1		/* O0 */,
	0x00, 1 	/* O1 */,
	0x00, 1		/* O2 */,
	0x04, 1		/* O3 */,
	0x08, 1		/* O4 */,
	0x0C, 1		/* O5 */,
	0x10, 1		/* O6 */,
	0x14, 1		/* O7 */,
	0x18, 1		/* O8 */,
	0x1C, 5		/* O9, O9.5 */,

	0x20, 5		/* B0, B0.5 */,
	0x24, 1		/* B1 */,
	0x28, 1		/* B2 */,
	0x2C, 1		/* B3 */,
	0x30, 1		/* B4 */,
	0x34, 1		/* B5 */,
	0x38, 1		/* B6 */,
	0x3C, 1		/* B7 */,
	0x40, 1		/* B8 */,
	0x44, 5		/* B9, B9.5 */,

	0x48, 1		/* A0 */,
	0x4c, 1		/* A1 */,
	0x50, 1		/* A2 */,
	0x54, 1		/* A3 */,
	0x58, 1		/* A4 */,
	0x5c, 1		/* A5 */,
	0x60, 1		/* A6 */,
	0x64, 1		/* A7 */,
	0x68, 1		/* A8 */,
	0x6c, 1		/* A9 */,

	0x70, 1		/* F0 */,
	0x72, 0		/* F1 (0/2)*/,
	0x74, 1		/* F2 */,
	0x78, 1		/* F3 */,
	0x7a, 0		/* F4 (3/5) */,
	0x7c, 1		/* F5 */,
	0x80, 1		/* F6 */,
	0x84, 1		/* F7 */,
	0x88, 1		/* F8 */,
	0x8a, 0		/* F9 (F8/G0) */,

	0x8c, 1		/* G0 */,
	0x90, 1		/* G1 */,
	0x94, 1		/* G2 */,
	0x98, 1		/* G3 */,
	0x9a, 0		/* G4 = G3/5 */,
	0x9c, 1		/* G5 */,
	0xa0, 1		/* G6 */,
	0xa2, 0		/* G7 = G6/8 */,
	0xa4, 1		/* G8 */,
	0xa6, 0		/* G9 = G8/K0 */,

	0xa8, 1		/* K0 */,
	0xac, 1		/* K1 */,
	0xb0, 1		/* K2 */,
	0xb4, 1 	/* K3 */,
	0xb8, 1		/* K4 */,
	0xbc, 1		/* K5 */,
	0xbe, 0		/* K6 = K5/M0 */,
	0xc0, 0		/* K7 = K5/M0 */,
	0xc2, 0		/* K8 = K5/M0 */,
	0xc3, 0		/* K9 = K5/M0 */,

	0xc4, 1		/* M0 */,
	0xc8, 1		/* M1 */,
	0xcc, 1		/* M2 */,
	0xd0, 1		/* M3 */,
	0xd4, 1		/* M4 */,
	0xd8, 1  	/* M5 */,
	0xdc, 1		/* M6 */,
	0xe0, 1		/* M7 */,
	0xe4, 1		/* M8 */,
	0xe8, 1		/* M9 */,
	0xec, 1		/* M10*/
	};
static unsigned char TABLE[] = 
  /*	 0         1         2         3         4         5         6  */
  /*	-012345678901234567890123456789012345678901234567890123456789012*/
	"?Ia0-IaIIIa-abII-IIIb-IIIab-bIII-IVaIVbIV-V-VI-VIImCNnnshpevarw";

static unsigned char *p, *e, c, sped[36];

/*===========================================================================
	Static variables for communications of internal routines
 *===========================================================================*/

static	char	err_level     = 1;	/* 0 = Warning	*/
static	char	display_error = 1;	/* Could be changed */

static struct {				/* Error report	*/
  unsigned char *str;	/* Text where error occured	*/
  unsigned char *end;	/* Indicates end of str		*/
  int  error;		/* Error level			*/
 } Str;			/* Error Structure 		*/

/*===========================================================================
		Error report
 *===========================================================================*/

static int error(msg, err)
/*++++++
.PURPOSE Display Error.
.RETURNS 0 (no error) / 1 (error)
--------*/
  char	*msg;	/* IN: Error Message */
  char	*err;	/* IN: Position of bad character */
{
  char *p, c;
    if (! msg)	return(0); 
    Str.error += err_level;
    if (display_error) {
    	c = *(Str.end), *(Str.end) = 0;
  	fprintf(stderr, "%s%s in: %s\n         ", 
  	    err_level ? "#***" : "#+++", msg, Str.str);
  	for (p = msg; *p; p++) fputc(' ', stderr); 
  	for (p = (char *)Str.str; p < err; p++) fputc(' ', stderr); 
  	fputc('^', stderr); fputc('\n', stderr);
    	*(Str.end) = c;
    }
    return(1);
}

/*===========================================================================*/
static int NextChar()
/*++++++
.PURPOSE Get next char (skip blanks), and translate () into :
.RETURNS The char, 0 when EOS;
.REMARKS 
--------*/
{
    while (p < e) {
    	c = *(p++);
  	if (isspace(c))	continue;
  	if (c == '(')	continue;
  	if (c == ')')	c = ':';
	/* else if (c == ',') c = '.';	Assume comma = dot (V1.41) */
  	return(c);
    }
    c = 0;
    return(0);
}

/*=======================================================================*/
static int match(cc)
/*++++++
.PURPOSE Reads until string differs
.RETURNS Number of matched chars
.REMARKS 
--------*/
  unsigned char *cc;	/* IN: Comparison String */
{
  unsigned char *q, u;
    for (q = cc; *q; q++) {
  	u = tolower(c);
  	if (u != *q)	break;
  	NextChar();
    }
    return(q-cc);
}

/*===========================================================================*/
static sp_class to_t()
/*++++++
.PURPOSE Converts the string to a temperature class.
.RETURNS The Spectral temperature
.REMARKS If an error is encountered, it is displayed
--------*/
{
  sp_class	tt, dt;
  int		i,j;
  unsigned char	*pt;
  struct TYPE2 	*ps;
	
    tt = ((sp_class)255)<<__SP_TEMP_;	/* Default Temperature	*/
    if (c == 0)	return(tt);	/* Empty String		*/

	/* Look first temperature without sub-class	*/
    for (ps = &spectra[0], i=ITEMS(spectra); --i>=0; ps++)  {
	if (ps->char1){			/* 2 letters */
	    if (c != ps->char1)	continue;
	    NextChar();
	}
  	for (j = ps->no, pt = ps->char2; j > 0; j-=2, pt += 2)
  		if (c == *pt)	goto Class_Found;
    }

    /* Spectral type doesn't match... */
    error("Invalid Temp. Class", p);
    return(tt);
	
  Class_Found: NextChar();
    tt = (int)(*++pt) << __SP_TEMP_;
  
	/* Get Subtype	*/
    if (isdigit(c))	{
  	i = (c-'0');
  	NextChar();
  	if (isdigit(c))	{ 
	    error("Invalid Temp. Subtype", p); 
	    return (((sp_class)255)<<__SP_TEMP_);
	}
    }
    else i = -1;

    if (ps->char2 == OBAFGKM) {		/* Normal Classification */
  	if (i >= 0) {	/* There is a sub--type (which can be bad!) */
  	    j = ((((int)ps->no - j)/2)*10 + i)*2;
  	    tt = TYPEN[j] << __SP_TEMP_;
  	    if (tt == 0) {
	    	error("Inexistant Temp. Subtype", p); 
		return (((sp_class)255)<<__SP_TEMP_);
	    }
  	    if (c == '.')	{	/* subsubclass	*/
  		dt = (TYPEN[j+2] - TYPEN[j]) << __SP_TEMP_;
  		NextChar();
  		if (isdigit(c)) {
  		    dt = dt * (c - '0') + (2<<__SP_TEMP_);
  		    tt += (dt / 10) & _SP_TEMP_;
  		    NextChar(c);
  		}
  	    }
  	}
  	else {
  	    tt = ((int)(pt[2] + pt[0])/2) << __SP_TEMP_;
  	    tt |= _SP_NOSUBCLASS_;
  	}
    }
    else if (i >= 0) {		/* Non-normal Spectral Type */
    	tt |= _SP_INCOMPLETE_;
  	while ((c == '.') || (c == ',') || (isdigit(c))) 
	    /* V1.41: Comma in C/S spectral type */
	    NextChar();
    }

    if (c == ':')	tt |= _SP_UTEMP_, NextChar();
  
    return(tt);
}

/*===========================================================================*/
static sp_class to_l(opt)
/*++++++
.PURPOSE Converts the string to luminosity class.
.RETURNS The Lumonosity Class
.REMARKS If an error is encountered, it is displayed
--------*/
  int opt;	/* IN: 0 = reset Luminosity class to 0 */
{
  sp_class	tt;
  int		i, x;
  static int 	lc;	/* Luminosity Class from previous call */

    tt = 0;	/* Default */
  
    i = lc;			/* LC from previous call */
    if ((opt == 0) || (c == 'I') || (c == 'V')) i = 0;

    while (c == 'I')	NextChar(),	i++;
    if (c == 'V')	NextChar(), 	i = 5 - i;
    while (c == 'I')	NextChar(),	i++;
    lc = i;		/* Keep LC for next call, e.g. IIab-b */

    if (i > 0) {	/* Check Possible ab */
  	if (i == 1)	x = 1, i = 6;	/* To make Ia / Ib different classes */
	else		x = 0, i = (i+1)<<2;
  	switch(tolower(c)) {
  	  case 'a':	i -= 1<<x;	break;
  	  case 'b':	i += 1<<x;	break;
  	  default:
  	  	if (x)	tt |= _SP_ULUM_;
	  	goto Done;
   	}
   	NextChar();

  	switch(tolower(c)) {
  	  case 'b':	i += 2;		break;
  	  case '-':	NextChar();
  	  	if (c != '0')	{ --p, c = '-'; goto Done; }
  	  case '0':	i -= 2;		break;
  	  default:
	  	goto Done;
   	}
   	NextChar();
    }

  Done:
    if (i > (7<<2)) {
	error("Invalid Lum. Class", p);
	i=0;
    }
    tt |= (i<<__SP_LUM_);
    if (c == ':')	tt |= _SP_ULUM_, NextChar();

    return(tt);
}
  
/*===========================================================================*/
int tr_aks(str, len, sp)
/*++++++
.PURPOSE Converts a string containing a spectral class to a long integer;
.RETURNS -1 for error, 0 for completely coded string, 1 if incomplete
.REMARKS If an error is encountered, message logged.
--------*/
  char	*str;	/* IN: String to scan	*/
  int	len;	/* IN: Length of str	*/
  sp_class *sp;	/* OUT: result		*/
{
  sp_class s, t1, t2, to_l(), to_t();
  unsigned char class;
  int		i;

    Str.str = p = (unsigned char *)str;
    Str.end = e = (unsigned char *)(str + len);
    Str.error = 0;

    s = ((sp_class)255)<<__SP_TEMP_;		/* Default Temperature	*/

	/* Check first the Average (X), Source (+), Quality (Number) */
    while ((c = NextChar())) {	
  	if (c == 'X')		{ s |= _SP_AVERAGED_; 	continue; }
  	if (c == '+')		{ s |= _SP_SOURCE_; 	continue; }
  	if (isdigit(c) && (c < '8'))	/* Quality	*/
  				{ s |= ((c-'0')<<8); 	continue; }
	break;
    }

	/* Check Now the (small letters) luminosity class */
    if (islower(c)) {
  	switch (c) {
	    case 'c':	/* Supergiant */
	    	s |= ( ( 6<<__SP_LUM_) | _SP_ULUM_);
	    	break;
	    case 'g':	/* Giant */
	    	s |= ( (16<<__SP_LUM_) | _SP_ULUM_);
	    	break;
	    case 'd':	/* Dwarf */
	    	s |= ( (24<<__SP_LUM_) | _SP_ULUM_);
	    	break;
	    case 's':	/* SubDwarf ? */
	    	NextChar();
		if (c == 'd')	s |=   (28<<__SP_LUM_);
	    	else	p -= 2;
	    	break;
	}
	NextChar();
	if (c == ':')	s |= _SP_ULUM_, NextChar();
    }

	/* Read the Temperature	*/
    class = c;
    t1 = to_t();
    s |= (t1 & ~_SP_TEMP_);	/* Take pecularities */
    if (c == '/')	s |= _SP_HOUK_;
  
    if ((c == '/') || (c == '-')) {/* Houk Coding: 2nd spectral type */
  	NextChar();
  	if (isdigit(c))	--p, c = class;
  	t2 = to_t();
  	s |= (t2 & ~_SP_TEMP_);
  	t1 >>= __SP_TEMP_, 	t2 >>= __SP_TEMP_;
		/* Check if not Too large temperature interval (except M0) */
  	i = (t2 == OBAFGKM[2*7-1] ? 8 : 4);   /* Max. temperature interval */
  	if ((t1 > t2) || ( (t2-t1) > i))
  		s |= (_SP_INCOMPLETE_|_SP_UTEMP_);
  	t1 = ((t1+t2)>>1) << __SP_TEMP_;
    }
    s = (s & ~_SP_TEMP_) | (t1 & _SP_TEMP_);
  
	/* Luminosity Classes */
    if ((c == 'I') || (c == 'V')) {
	t1 = to_l(0);
	if (c == '/')	s |= _SP_HOUK_;
	if ((c == '-') || (c == '/'))	{
		NextChar();
		t2 = to_l(1);
  		i  = (t1+t2)>>1;
  		t1 = ((t1|t2) & ~_SP_LUM_) | (i & _SP_LUM_);
	}
  	s = (s & ~_SP_LUM_) | t1;
    }
  
	/* Pecularities */

    while(c) {
	switch(tolower(c)) {
	  case '+':	s |= _SP_COMPOSITE_;
	  	break;
	  case 'e':	s |= _SP_EMISSION_;
	  	break;
	  case 'm':	s |= _SP_M_;
	  	break;
	  case 'w':	s |= _SP_W_;
	  	break;
	  case 's':	NextChar();
	  	if (tolower(c) != 'h')	{ s |= _SP_S_; continue; }
	  	s |= _SP_SHELL_; 	break;
	  case 'n':	NextChar();
	  	if (tolower(c) != 'n')	{ s |= _SP_N_; continue; }
	  	s |= _SP_NN_; 		break;
	  case 'c':	NextChar();
	  	if (tolower(c) == 'n')
	  	{ s |= _SP_CN_; 	break; }
	  	if (match("comp") > 2)
	  		s |= _SP_COMPOSITE_;
	  	else 	s |= _SP_INCOMPLETE_; 
	  	continue; 
	  case 'p':	s |= _SP_PECULIAR_;
		if (match("peculiar") == 2)	s |= _SP_EMISSION_;
	  	continue;
	  case 'v':	s |= _SP_VARIABLE_;
		match("variable");
	  	continue;
	  default: 	s |= _SP_INCOMPLETE_; 
	}
	NextChar();
    }

    *sp = s;
    if (Str.error) {
  	*sp |= _SP_INCOMPLETE_;
  	i = -1;
    }
    else i = (s & _SP_INCOMPLETE_ ? 1 : 0);
  
    return(i)  ;
}

/*===========================================================================*/
static int ed_t(tt)
/*++++++
.PURPOSE Edit temperature class
.RETURNS Length of edited string
.REMARKS 
--------*/
  sp_class	tt;	/* IN: Temperature Class */
{
  int		it, i, k;
  unsigned char	dt;
  struct TYPE2 	*ps;
  static unsigned char	SSTYPE[]= {'2', '5', '7'};
	
    it = SpTemp(tt);

    for (i=3; i<sizeof(OBAFGKM); i+=2)
  	if (it < (int)OBAFGKM[i])	break;
    i -= 3;
  
    if ((it < 0xff) && (it >= 0xf0 )) {	/* 2 letters */
  	for (ps = &spectra[2]; ps < &spectra[ITEMS(spectra)]; ps++) {
  		for (i=1; i<(int)ps->no; i += 2)
  			if (ps->char2[i] == it)	break;
  		if (i < (int)ps->no)	break;
  	}
  	*(p++) = ps->char1;
  	*(p++) = ps->char2[--i];
  	goto FIN;
    }

	/* Normal (OBAFGKM) Spectral Type */
    *(p++) = OBAFGKM[i];
    if (i >= (int)spectra[0].no)goto FIN;
    if (tt & _SP_NOSUBCLASS_)	goto FIN;

    i *= 10;
    while(it >= TYPEN[i]) i += 2;
				/* Find Main Class i = [OBAFGKM]*10 + subtype */
    if(TYPEN[i] > it) 	i -= 2;
  
    dt = it - TYPEN[i];
    if (tt & _SP_HOUK_)	k = TYPEN[i+1] & (1 << dt);
    else			k = 1;

    if (k) {		/* Normal Type */
  	*(p++) = '0' + (i>>1)%10;
  	if (dt)
  	{	*(p++) = '.';
  		if ((TYPEN[i+2] - TYPEN[i]) < 3) dt = 2;
  		*(p++) = SSTYPE[dt-1];
  	}
  	goto FIN;
    }

			/* Houk's Classification */
    --p;
    if ( (dt&1) && (TYPEN[i+1] & 4)) {	/* .5 subclass exists */
  	ed_t(tt-(1<<__SP_TEMP_)), *(p++) = '/';
  	ed_t(tt+(1<<__SP_TEMP_));
  	goto FIN;
    }
  
    for (k=i+1; TYPEN[k] == 0; k -= 2)	;	/* Left Type	*/
    ed_t(TYPEN[k-1] << __SP_TEMP_);
    *(p++) = '/';
    for (k=i+3; TYPEN[k] == 0; k += 2)	;	/* Right Type 	*/
    ed_t(TYPEN[k-1] << __SP_TEMP_);

  FIN:
    return(p - sped);
}

/*===========================================================================*/
static int edx(n, tt)
/*++++++
.PURPOSE Edit number according to table tt
.RETURNS 0
.REMARKS 
--------*/
  int		n;	/* IN: Number */
  unsigned char	*tt;	/* IN: Table  */
{
  int	it, j;
  unsigned char  *q;
	
    if (!n) return(0);
    it = n;
    while (it) {
   	it = (it-1)<<2;
   	j = (c ? tt[it+1] : tt[it]);
	for (q = &TABLE[tt[it+2]]; --j >= 0; p++, q++) *p = *q;
   	it = (c ? tt[it+3] : 0);
   	*(p++) = '/';
    }
    --p;
    return(0);
}

/*===========================================================================*/
static int ed_l(tt)
/*++++++
.PURPOSE Edit Luminosity class
.RETURNS 
.REMARKS 
--------*/
  sp_class	tt;	/* IN: Temperature Class */
{
  int		it;
  static unsigned char DECOL[] = {
          1,1,0,0,		
          3,3,1,0,		/* Ia0 	*/
          6,3,1,4,
          2,2,1,0,		/* Ia	*/
          5,2,9,6,
          3,3,24,0,		/* Iab	*/
          5,3,24,8,
          2,2,19,0,		/* Ib	*/
          5,2,19,12,
          5,2,19,12,
          3,2,8,12,
          2,2,8,0,		/* II	*/
          3,2,18,16,
          6,2,14,16,
          4,2,7,16,
          3,3,7,0,		/* III	*/
          4,3,17,20,
          6,3,29,20,
          3,3,33,20,
          2,2,33,0,		/* IV	*/
          2,3,33,24,
          4,2,39,24,
          2,1,34,24,
          1,1,34,0,		/* V	*/
          2,1,37,0,
          4,1,42,28,
          4,2,42,28,
          2,2,44,0,
          2,2,44,0,		/* sd	*/
          6,2,44,32,
          3,3,47,0,
          3,3,47,0
	} ;

   it = (SpLum(tt)); 
   edx(it, DECOL);
   return(0);
}

/*===========================================================================*/
char *kstoa(tt)
/*++++++
.PURPOSE Edit a Coded Spectral class
.RETURNS Address of edited Spectral Class (terminated with an EOS)
.REMARKS
--------*/
  sp_class	tt;	/* IN: Spectral Class */
{
  int		i, j;
  sp_class	t;
  static unsigned char pecularities[] = {
          2,2,51,0, 	/* CN	*/
          1,1,62,0, 	/* w	*/
          1,1,50,0, 	/* m	*/
          1,1,55,0, 	/* s	*/
          1,1,53,0, 	/* n	*/
          2,2,53,0, 	/* nn	*/
          1,1,57,0, 	/* p	*/
          2,2,55,0, 	/* sh	*/
          1,1,58,0, 	/* e	*/
          3,3,59,0 	/* var	*/
  };

    p = sped;

    c = (tt & _SP_HOUK_ ? 1 : 0);
    if (tt & _SP_AVERAGED_)	*(p++) = 'X';
    if (tt & _SP_SOURCE_)	*(p++) = '+';
    if ((i = (tt & _SP_QUAL_)))	*(p++) = '0' + (i>>__SP_QUAL_);
  
    if (p != sped)	*(p++) = ' ';
  
	/* Subdwarfs */

    if ( (tt & _SP_LUM_) == (28 << __SP_LUM_)) {
  	*(p++) = 's', *(p++) = 'd', tt &= ~_SP_LUM_;
	if (tt & _SP_ULUM_)	*(p++) = ':';
	tt &= ~_SP_ULUM_;
    }

    ed_t(tt);		/* Edit temperature	*/
    if (tt & _SP_UTEMP_)	*(p++) = ':';
  
    ed_l(tt);		/* Edit Luminosity	*/
    if (tt & _SP_ULUM_)	*(p++) = ':';

	/* Edit Pecularities */
    t = tt;
    for (i=0, j=2; --j >= 0; t >>= 2, i += 12)
  	edx(t&3, &pecularities[i]);
    for (j=4; --j >= 0; t >>= 1, i += 4)
  	if (t&1) edx(1, &pecularities[i]);

    if (tt & _SP_COMPOSITE_)	*(p++) = '+';
    if (tt & _SP_INCOMPLETE_)	*(p++) = '.', *(p++) = '.', *(p++) = '.';
  
    *p = '\0';
  
    return((char *)sped);
}

/*===========================================================================*/
int ed_ks(str, size, tt)
/*++++++
.PURPOSE Edit a Coded Spectral class
.RETURNS Length of edited string (may be up to 36 bytes)
.REMARKS If str is too small, edition is truncated. But the returned
	length is still the required length.
--------*/
  char	*str;	/* OUT: Output string	*/
  int	size;	/* IN: Size of str	*/
  sp_class tt;	/* IN: Spectral Class */
{
  int	l;
	
    kstoa(tt);		/* Edit in local buffer	*/
    l = (p - sped) + 1;	/* Length of output, including EOS	*/
    l = MIN(l, size);
    strncpy(str, (char *)sped, l);
  
    return(l-1);
}

#ifdef TEST
#include <stdlib.h>
#include <unistd.h>

static int posdata[2] = {-1,-1};
static int posbyte[2] = {-1,-1};
static char sepchar;		/* -d option */
static char addchar;		/* -a option */
static char optr;		/* -r option */
static char optt;		/* -t option */
static char usage[] = "\
Usage: trsp [-h] [-version] [-ac] [-dc] [-c.] [-r] [-t] [file]";
static char help[] = "\
  -h: print this help message and exit\n\
 -ac: append the delimiter 'c' before the coded spectral type\n\
 -dc: define 'c' as the delimiter character (-d alone: tab)\n\
 -c.: define the column containing the spectral type (default: all)\n\
  -r: replace the spectral type by its coded value;\n\
  -t: input is a table (data start after the --- delimiter)\n\
 file: input file (default is stdin)\n\
";

static int set_posdata(char *str, int *posdata)
/*++++++++++
.PURPOSE Interpret the "-c" argument
.RETURNS Number of bytes interpreted.
-----------*/
{
  char *p; int n ;

    for (p=str, n=0; isdigit(*p); p++) n = n*10 + (*p - '0') ;
    if (n <= 0) return(0) ;
    posdata[0] = n -1;
    /*if (*p != '-') return(p-str) ; */
    if (ispunct(*p)) 
    	   for (++p, n=0; isdigit(*p); p++) n = n*10 + (*p - '0') ;
    posdata[1] = n ;
   return(p-str) ;
}

static int set_bytes(int *byte, int *field, char *text)
/*++++++++++
.PURPOSE Transform a field position into a byte position
.RETURNS Number of fields involved (field[1] - field[0])
.REMARKS Set to -1 for field missing
  byte (OUT): position in bytes (from 0)
  field (IN): position in field (from 0) 
  text  (IN): String containing the fields 
-----------*/
{
  char *p; int k;

    byte[0] = byte[1] = -1;	/* Not here */
    if ((field[0]<0) || (field[1]<field[0]))
	return(-1); 

    /* Locate position of field[0] */
    for (k=0, p=text; k<field[0]; k++) {
	if ((p = strchr(p, sepchar))) p++;
	else return(-1);	/* Not here */
    }

    byte[0] = p-text;
    /* Locate now the other fields */
    while (1) {
	while (*p && (*p != sepchar)) p++;
	if (!*p) break;		/* EOS found */
	if (++k >= field[1]) break;
	p++;			/* skip TAB  */
    }
    while (*p && (*p != sepchar)) p++;
    byte[1] =  p-text;

    return( byte[1] - byte[0]);
}

int main(int argc, char **argv)
/*++++++++++++++++
.PURPOSE  TEST PROGRAM
.RETURNS  to OS
-----------------*/
{ 
  char abuf[BUFSIZ];
  sp_class sp;
  FILE *input=0;
  int 	i, n, phase=1;	/* even = in header, odd = in data */
  char *p;

    while (-- argc > 0) {
	 p = *++argv;
	 if (*p == '-') switch(*++p) {
	     case 'h':  printf("%s\n%s", usage, help); exit(0);
	     case 'v':	/* Assume -version */
		printf("trsp (CDS) Version %s\n", VERSION);
		exit(0);
	     case 'c': 
		p++;
		i = set_posdata(p, posdata); 
		if (!p[i]) continue ;
		fprintf(stderr, "\n#***Bad argument: '%s'\n", *argv);
		exit(1) ;
	     case 'd':
		sepchar = p[1];
		if (!sepchar) sepchar = '\t';
		continue;
	     case 'a':
		addchar = p[1];
		if (!addchar) addchar = '\t';
		continue;
	     case 'r':
		optr = 1;
		continue;
	     case 't':
		optt = 1; phase = 0;
		continue;
	     default: 
		fprintf(stderr, "\n#***Bad option: -%s\n", p);
		fprintf(stderr, "%s\n", usage);
		exit(1);
  	}
	else if (input) {
	    fprintf(stderr, "#***Only one input file!\n%s\n", usage);
	    exit(1);
	}
	input = fopen(p, "r");
	if (!input) {
	    fprintf(stderr, "#***trsp(%s): ", p);
	    perror("");
	    exit(1);
	}
    }
    if (!input) input = stdin;

    /* Appending char */
    if (optr) addchar = 0;
    else if (addchar == 0) addchar = sepchar ? sepchar : '\t' ;
  
    for (;;) { 
	if (isatty(0)) printf("Input Class: ");
    	if (!fgets(abuf, sizeof(abuf), input)) break;
	if (abuf[0] == '%')   { printf("%s", abuf) ; continue ; }
	if (abuf[0] == '#')   { printf("%s", abuf) ; continue ; }
	if (abuf[0] == '!')   { printf("%s", abuf) ; continue ; }
	/* Remove tail spaces */
	for (i = strlen(abuf)-1; (i>=0) && isspace(abuf[i]); i--) abuf[i] = 0;
	if (phase==0) {	/* -t option: must find the data part */
	    for (i=0; (abuf[i]=='-')||(abuf[i]==sepchar); i++) {
		if (abuf[i] == 0) break;
	    }
	    if ((i>0) && (abuf[i]==0)) phase = 1;
	    printf("%s\n", abuf);
	    continue;
	}
	if (posdata[0] >= 0) {	/* Data in dedicated Columns 	*/
	    /* V4.9: char-delimited columns => convert posdata into posbyte */
	    if (sepchar) for (i=0; i<ITEMS(posdata); i += 2) {
		if (posdata[i]>=0) {
		    if (set_bytes(posbyte+i, posdata+i, abuf)<0) {
			/* If position not there ==> may be a comment! */
			if (i == 0) break;
		    }
		}
		else { posbyte[i] = posdata[i]; posbyte[i+1] = posdata[i+1]; }
	    }
	    else memcpy(posbyte, posdata, sizeof(posdata));
	    if (posbyte[0]<0) { printf("%s\n", abuf) ; continue ; }  /* No sp */
	    p = abuf+posbyte[0]; 
	    n = posbyte[1] - posbyte[0]; 
	}
	else { 
	    p = abuf;
	    n = strlen(abuf);
	}
    	i = tr_aks(p, n, &sp);

	/* Write the result */
        if (optr) 		/* Replace current coordinates */
	    fwrite(abuf, 1, posbyte[0], stdout);
	else printf("%s%c", abuf, addchar);
	printf("%03d.%02d:%d:%02X", SpTemp(sp), SpLum(sp), SpQual(sp), SpPec(sp));
	if (optr) printf("%s", abuf+posbyte[1]);
	putchar('\n');
	/*printf("---->%08X (%d)\n", sp, i);*/
    	/*printf(kstoa(sp));*/
    }
    exit(0);
}  
#endif
